''' opencos.eda_config - handles --config-yml arg, and providing a
'config' dict for use by opencos.eda, opencos.commands, and opencos.tools

Order of precedence for default value of eda arg: --config-yaml:
1) os.environ.get('EDA_CONFIG_YML', '') -- filepath to an eda_config.yml file
2) ~/.opencos-eda/EDA_CONFIG.yml
3) (package pip installed dir for opencos)/eda_config_defaults.yml
'''

import copy
import os
import argparse

import mergedeep

from opencos import util
from opencos.files import safe_shutil_which
from opencos.util import safe_emoji
from opencos.utils.markup_helpers import yaml_safe_load, yaml_safe_writer

class Defaults:
    '''Defaults is a global namespace for constants and supported features.

    Defaults.config_yml is set depending on search order for default eda_config[_defaults].yml

    Note that opencos.eda and other packages are free to set and add additonal keys to the
    "config", but the "supported_config_keys" are only what's allowed in the initial
    --config-yml=YAML_FILE

    Note that for any supported*keys str, we will allow name_with_underscores or
    name-with-dashes. If you have both, you will get an error.
    '''

    environ_config_yml = ''
    home_config_yml = ''
    opencos_config_yml = 'eda_config_defaults.yml'
    config_yml = ''
    config_yml_set_from = ''
    non_default_config_yml_arg_used = ''

    # partial cfg, merge w/ defaults
    environ_partial_config_yml = ''
    home_partial_config_yml = ''
    partial_config_yml = ''
    partial_config_yml_set_from = ''

    supported_config_keys = set([
        'DEFAULT_HANDLERS', 'DEFAULT_HANDLERS_HELP',
        'defines',
        'dep_command_enables',
        'dep_tags_enables',
        'deps_markup_supported',
        'deps_subprocess_shell',
        'bare_plusarg_supported',
        'deps_expandvars_enable',
        'show_tool_versions',
        'dep_sub',
        'vars',
        'file_extensions',
        'use_max_threads',
        'command_loads_all_tools',
        'command_determines_tool',
        'command_tool_is_optional',
        'command_available_with_tool_if_tool_not_installed',
        'command_has_subcommands',
        'tools',
        'auto_tools_order',
    ])
    supported_config_tool_keys = set([
        'exe',
        'macos-app-name', 'macos-prefer-exes',
        'windows-app-name', 'windows-prefer-exes',
        'app-name',
        'docker-support', 'docker-default-image', 'docker-env',
        'handlers',
        'requires-env', 'requires-py', 'requires-cmd',
        'requires-vsim-helper',
        'requires-vscode-extension',
        'requires-docker',
        'preserve-other-tool-config',
        'disable-tools-multi', 'disable-auto',
        'defines',
        'log-bad-strings',
        'log-must-strings',
        'log-warning-strings',
        'sim-libraries',
        'compile-args',
        'compile-waves-args',
        'compile-waivers',
        'compile-coverage-args',
        'elab-args',
        'elab-waves-args',
        'simulate-args',
        'simulate-waves-args',
        'simulate-waivers',
        'simulate-coverage-tcl',
        'coverage-args',
        'tcl-command-args',
        'waves-file-ext', 'upload-file-ext',
    ])


EDA_OUTPUT_CONFIG_FNAME = 'eda_output_config.yml'


def set_defaults() -> None:
    '''Updates Defaults *config_yml members, sets Defaults.config_yml

    This is for user supplied config that completely overrides the default provided by the
    pypackage (eda_config_defaults.yml)
    '''

    Defaults.environ_config_yml = os.environ.get(
        'EDA_CONFIG_YML', os.environ.get('EDA_CONFIG_YAML', '')
    )
    if Defaults.environ_config_yml and \
       os.path.isfile(Defaults.environ_config_yml):
        Defaults.config_yml = Defaults.environ_config_yml
        Defaults.config_yml_set_from = 'env EDA_CONFIG_YML'
        return

    home = os.environ.get('HOME', os.environ.get('HOMEPATH', ''))
    if home and os.path.isdir(os.path.join(home, '.opencos-eda')):
        Defaults.home_config_yml = [
            os.path.join(home, '.opencos-eda', 'eda_config_defaults.yml'),
            os.path.join(home, '.opencos-eda', 'eda_config_defaults.yaml'),
            os.path.join(home, '.opencos-eda', 'EDA_CONFIG.yml'),
            os.path.join(home, '.opencos-eda', 'EDA_CONFIG.yaml')
        ]
        for x in Defaults.home_config_yml:
            if os.path.isfile(x):
                Defaults.config_yml = x
                Defaults.config_yml_set_from = f'file [$HOME|$HOMEPATH]/.opencos-eda/{x}'
                return

    # else default:
    Defaults.config_yml = Defaults.opencos_config_yml
    Defaults.config_yml_set_from = ''


def set_partial_config_defaults() -> None:
    '''Updates Defaults *partial_config_yml members, sets Defaults.partial_config_yml

    This is for partial configs that will be merged into the default config
    '''

    Defaults.environ_partial_config_yml = os.environ.get(
        'EDA_PARTIAL_CONFIG_YML', os.environ.get('EDA_PARTIAL_CONFIG_YAML', '')
    )
    if Defaults.environ_partial_config_yml and \
       os.path.isfile(Defaults.environ_partial_config_yml):
        Defaults.partial_config_yml = Defaults.environ_partial_config_yml
        Defaults.partial_config_yml_set_from = 'env EDA_PARTIAL_CONFIG_YML'
        return

    home = os.environ.get('HOME', os.environ.get('HOMEPATH', ''))
    if home and os.path.isdir(os.path.join(home, '.opencos-eda')):
        Defaults.home_partial_config_yml = [
            os.path.join(home, '.opencos-eda', 'eda_partial_config.yml'),
            os.path.join(home, '.opencos-eda', 'eda_partial_config.yaml'),
        ]
        for x in Defaults.home_partial_config_yml:
            if os.path.isfile(x):
                Defaults.partial_config_yml = x
                Defaults.partial_config_yml_set_from = f'file [$HOME|$HOMEPATH]/.opencos-eda/{x}'
                return

    # else default:
    Defaults.partial_config_yml = ''
    Defaults.partial_config_yml_set_from = ''


set_defaults()
set_partial_config_defaults()


def find_eda_config_yml_fpath(
        filename:str, package_search_only=False, package_search_enabled=True
) -> str:
    '''Locates the filename (.yml) either from fullpath provided or from the sys.path
    (pip-installed) opencos-eda package paths.'''

    # Check fullpath, unless we're only checking the installed pacakge dir.
    if package_search_only:
        pass
    elif os.path.exists(filename):
        return os.path.abspath(filename)

    leaf_filename = os.path.split(filename)[1]

    if leaf_filename != filename:
        # filename had subdirs, and we didn't find it already.
        util.error(f'eda_config: Could not find {filename=}')
        return None

    # Search in . or pacakge installed dir
    thispath = os.path.dirname(__file__) # this is not an executable, should be in packages dir.

    if package_search_only:
        paths = [thispath]
    elif package_search_enabled:
        paths = ['', thispath]
    else:
        paths = ['']


    for dpath in paths:
        fpath = os.path.join(dpath, leaf_filename)
        if os.path.exists(fpath):
            return fpath

    util.error(f'eda_config: Could not find {leaf_filename=} in opencos within {paths=}')
    return None


def check_config(config:dict, filename='') -> None:
    '''Returns None, will util.error(..) if there are issues in 'config'

    checks for known dict keys and data types, is NOT exhaustive checking.
    Also will attempt to patch incorrectly named keys (underscore vs dash) in
    name with only a warning.
    '''

    # sanity checks:
    if not config:
        util.error(
            f'eda_config.get_config({filename=}): is empty'
        )
        return

    for key in list(config.keys()):
        if key not in Defaults.supported_config_keys:
            # For these, we prefer names with underscores. If the name has a '-', replace
            # with underscore.
            alt_key = key.replace('-', '_')
            if alt_key in Defaults.supported_config_keys and alt_key not in config:
                util.warning(
                    f'eda_config.get_config({filename=}): has unsupported {key=}, replacing ',
                    f'with underscores: {alt_key}'
                )
                config[alt_key] = copy.deepcopy(config[key])
                del config[key] # safe, iterator is key in (list copy of config keys)
            else:
                util.error(
                    f'eda_config.get_config({filename=}): has unsupported {key=},',
                    f' {Defaults.supported_config_keys=}'
                )

    for tool, table in config.get('tools', {}).items():
        for key in list(table.keys()):
            if key not in Defaults.supported_config_tool_keys:
                # For these, we prefer names with dashes. If the name has a '_', replace with dash
                alt_key = key.replace('_', '-')
                if alt_key in Defaults.supported_config_keys and alt_key not in config:
                    util.warning(
                        f'eda_config.get_config({filename=}): has unsupported {key=} in',
                        f'tools.{tool}, replacing with: {alt_key}'
                    )
                    config[tool][alt_key] = copy.deepcopy(table[key])
                    del table[key] # safe, iterator is key in (list copy of table keys)
                else:
                    util.error(
                        f'eda_config.get_config({filename=}): has unsupported {key=} in',
                        f'tools.{tool=}, {Defaults.supported_config_tool_keys=}'
                    )


def update_config_auto_tool_order_for_tool(tool: str, config: dict) -> str:
    '''Update config entry if the value for tool is in the form 'name=/path/to/exe

    Input arg tool can be in the form (for example):
      tool='verlator', tool='verilator=/path/to/verilator.exe'

    Performs no update if tool has no = or : in it. Returns tool (str) w/out = in it
    '''
    return tool_try_add_to_path(tool=tool, config=config, update_config=True)


def update_config_auto_tool_order_for_tools(tools: list, config: dict) -> list:
    '''Given a list of tools and eda_config style 'config' dict, update

    the auto_tool_order (consumed by opencos.eda when --tool is not specified).
    '''
    ret = []
    for tool in tools:
        ret.append(update_config_auto_tool_order_for_tool(tool, config))
    return ret


def update_config_for_eda_safe(config) -> None:
    '''Set method to update config dict values to run in a "safe" mode'''
    config['dep_command_enables']['shell'] = False


def deps_shell_commands_enabled(config) -> bool:
    '''Get method on config to determine if DEPS.yml shell-style commands are allowed'''
    return config['dep_command_enables']['shell']


def get_config(filename) -> dict:
    '''Given an eda_config_default.yml (or --config-yml=<filename>) return a config

    dict from the filename.'''

    fpath = find_eda_config_yml_fpath(filename)
    user_config = yaml_safe_load(fpath)
    check_config(user_config, filename=filename)

    # The final thing we do is update key 'config-yml' with the full path used.
    # This way we don't have to pass around --config-yml as some special arg
    # in eda.CommandDesign.args, and eda.CommandMulti can use when re-invoking 'eda'.
    user_config['config-yml'] = fpath
    return user_config


def get_config_handle_defaults(filename) -> dict:
    '''Given a user provided --config-yml=<filename>, return a merged config with

    the existing default config.'''

    user_config = get_config(filename)
    user_config = get_config_merged_with_defaults(user_config)
    return user_config


def merge_config(dst_config:dict, overrides_config:dict, additive_strategy=False) -> None:
    '''Mutates dst_config, uses Strategy.TYPESAFE_REPLACE'''
    # TODO(drew): It would be cool if I could have Sets be additive, but oh well,
    # this gives the user more control over replacing entire lists.
    strategy = mergedeep.Strategy.TYPESAFE_REPLACE
    if additive_strategy:
        strategy = mergedeep.Strategy.TYPESAFE_ADDITIVE
    mergedeep.merge(dst_config, overrides_config, strategy=strategy)


def get_config_merged_with_defaults(config: dict) -> dict:
    '''Returns a new config that has been merged with the default config.

    The default config location is based on Defaults.config_yml (env, local, or pip
    installed location)

    This will also include merging with Defaults.partial_config_yml (env, home) dir
    '''

    if Defaults.config_yml_set_from:
        # use "parsed.config_yml" from the argparser
        default_fpath = Defaults.config_yml
    else:
        # use default (have to find it in package/dist)
        default_fpath = find_eda_config_yml_fpath(Defaults.config_yml, package_search_only=True)
    default_config = yaml_safe_load(default_fpath)
    merge_config(default_config, overrides_config=config)

    if Defaults.partial_config_yml:
        # user partial config
        partial_config = yaml_safe_load(Defaults.partial_config_yml)
        merge_config(default_config, overrides_config=partial_config)

    # This technically mutated updated into default_config, so return that one
    return default_config


def get_argparser() -> argparse.ArgumentParser:
    '''Returns an ArgumentParser, handles --config-yml=<filename> arg'''

    # re-run set_defaults() in case a --env-file overwrote Defaults.config_yml
    set_defaults()

    parser = argparse.ArgumentParser(
        prog=f'{safe_emoji("🔎 ")}opencos eda config options', add_help=False, allow_abbrev=False
    )
    parser.add_argument(
        '--config-yml', type=str, default=Defaults.config_yml,
        help=(
            f'YAML filename to use for configuration (default {Defaults.config_yml}).'
            ' Can be overriden using environmnet var EDA_CONFIG_YML=FILE, or from file'
            ' [$HOME|$HOMEPATH]/.opencos-eda/EDA_CONFIG.[yml|yaml] or'
            ' [$HOME|$HOMEPATH]/.opencos-eda/eda_config_defaults.[yml|yaml]'
        )
    )
    parser.add_argument(
        '--partial-config-yml', type=str, default=Defaults.partial_config_yml,
        help=(
            'YAML filename to use for partial configuration.'
            ' Can be overriden using environmnet var EDA_PARTIAL_CONFIG_YML=FILE, or from file'
            ' [$HOME|$HOMEPATH]/.opencos-eda/eda_partial_config.[yml|yaml]'
        )
    )
    return parser


def get_argparser_short_help() -> str:
    '''Returns a shortened help string given for arg --config-yml.'''
    return util.get_argparser_short_help(parser=get_argparser())


def get_eda_config(args:list, quiet=False) -> (dict, list): # pylint: disable=too-many-branches
    '''Returns an config dict and a list of args to be passed downstream
    to eda.main and eda.process_tokens.

    Handles args for:
      --config-yml=<YAMLFILE> --partial-config-yml=<YAMLFILE>

    This will merge the result with the default config (if overriden)
    '''
    parser = get_argparser()
    try:
        parsed, unparsed = parser.parse_known_args(args + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        util.error(f'problem attempting to parse_known_args for {args=}')

    util.debug(f'eda_config.get_eda_config: {parsed=} {unparsed=}  from {args=}')

    if parsed.config_yml:
        if not quiet:
            if parsed.config_yml != Defaults.config_yml:
                # It was set on CLI:
                util.info(f'eda_config: --config-yml={parsed.config_yml} observed')
            elif Defaults.config_yml_set_from:
                # It was picked up via env or HOME/.opencos-eda/ override:
                util.info(f'eda_config: --config-yml={parsed.config_yml} observed, from',
                          f'{Defaults.config_yml_set_from}')
        fullpath = find_eda_config_yml_fpath(parsed.config_yml)
        config = get_config(fullpath)
        if not quiet:
            util.info(f'eda_config: using config: {fullpath}')
        # Calling get_config(fullpath) will add fullpath to config['config-yml'], so the
        # arg for --config-yml does not need to be re-added.
    else:
        config = None

    if parsed.partial_config_yml:
        if not quiet:
            if Defaults.partial_config_yml != parsed.partial_config_yml:
                Defaults.partial_config_yml_set_from = 'CLI args'
            # We do not have to "find" this one, it's already found, just use it
            Defaults.partial_config_yml = parsed.partial_config_yml
            util.info(f'eda_config: using parital config: {Defaults.partial_config_yml}')


    do_merge = False
    if parsed.config_yml != Defaults.config_yml:
        Defaults.non_default_config_yml_arg_used = parsed.config_yml
        do_merge = True
    if parsed.partial_config_yml:
        do_merge = True

    if do_merge:
        # This will handle both Defaults.config_yml and Defaults.partial_config_yml:
        config = get_config_merged_with_defaults(config)

    return config, unparsed


def get_config_yml_args_for_flist() -> list:
    '''Returns list of args, or empty list. Used by CommandFList when we want to get the args

    to reproduce a target to be run again in `eda`'''
    ret = []

    if Defaults.non_default_config_yml_arg_used:
        ret.append(f'--config-yml={Defaults.non_default_config_yml_arg_used}')

        if Defaults.config_yml_set_from:
            # a default config-yml was used, but it wasn't from eda, was from
            # ENV or HOME dir. This is not included in the flist, so warn about it.
            # Since we don't support > 1 config-yml args, need to warn about this.
            util.warning('Note: for command "flist", picked up',
                         f'--config-yml={Defaults.config_yml},',
                         f'from: {Defaults.config_yml_set_from}')

    elif Defaults.config_yml_set_from:
        ret.append(f'--config-yml={Defaults.config_yml}')

    return ret


def write_eda_config_and_args(
        dirpath : str, filename: str = EDA_OUTPUT_CONFIG_FNAME,
        command_obj_ref: object = None
) -> None:
    '''Writes and eda_config style dict to dirpath/filename'''
    if command_obj_ref is None:
        return
    fullpath = os.path.join(dirpath, filename)
    data = {}
    for x in ['command_name', 'config', 'target', 'args', 'modified_args', 'defines',
              'incdirs', 'files_v', 'files_sv', 'files_vhd', 'files_cpp', 'files_sdc',
              'files_non_source']:
        # Use deep copy b/c otherwise these are references to opencos.eda.
        data[x] = copy.deepcopy(getattr(command_obj_ref, x, ''))

    # copy util.args, and other util globals:
    data['util'] = {}
    for x in ['INITIAL_CWD', 'args', 'dot_f_files_expanded', 'env_files_loaded', 'max_error_code']:
        data['util'][x] = getattr(util, x, '')

    # copy some information about which eda_config YAML was used:
    for member in ['config_yml', 'config_yml_set_from', 'non_default_config_yml_arg_used']:
        data[member] = getattr(Defaults, member, '')

    # fix some burried class references in command_obj_ref.config,
    # otherwise we won't be able to safe load this yaml, so cast as str repr.
    config = getattr(command_obj_ref, 'config', {})
    for k, v in config.items():
        if k == 'command_handler':
            data['config']['command_handler'] = str(v)


    yaml_safe_writer(data=data, filepath=fullpath)
    util.artifacts.add(name=fullpath, typ='yaml', description='opencos-eda output config YAML')



def tool_arg_get_parts(tool: str) -> list:
    '''Given a tool (str or None) that may be in form <name>=/path/to/something

    Return the parts [<name>, <path>, ..]
    '''
    if not tool or ('=' not in tool and ':' not in tool):
        return [tool]

    if '=' in tool:
        parts = tool.split('=')
    else:
        parts = tool.split(':')

    return parts

def tool_arg_remove_path_information(tool: str) -> str:
    '''Given a tool (str or None) that may be in form <name>=/path/to/something

    Return the <name> only
    '''
    if not tool:
        return tool
    return tool_arg_get_parts(tool)[0]


def tool_get_exe_list(tool: str, config: dict) -> list:
    '''Gets the exe (list) from config.tools.TOOL.exe'''

    exe = config['tools'][tool].get('exe', '')
    if not exe:
        # exe not present assume 'tool' name is the exe.
        return [tool] # make it a list
    if isinstance(exe, list):
        return exe # already a list

    # assume single str item
    return [exe] # make it a list


def tool_get_first_exe(tool: str, config: dict) -> str:
    '''Gets the first exe from config.tools.TOOL.exe'''

    exe = tool_get_exe_list(tool, config)[0]
    assert isinstance(exe, str), f'config.tools.{tool}.exe - return should be a str or list-of-str'
    return exe


def tool_try_add_to_path( # pylint: disable=too-many-branches
        tool: str, config: dict, update_config: bool
) -> str:
    '''Since we support --tool=<name>=/path/to/bin[/exe], attempt to prepend $PATH

    (also works for --tool=<name>:/path/to/bin[/exe] )

    with this information for this tool (which will nicely affect all subprocesses,
    but not wreck our original shell).'''

    name_path_parts = tool_arg_get_parts(tool)
    if len(name_path_parts) == 1:
        return name_path_parts[0]

    name, path_arg = name_path_parts[0:2]

    if name not in config['tools']:
        return name

    config_exe = tool_get_first_exe(tool=name, config=config)

    if path_arg and os.path.isfile(path_arg):
        # Someone passes us --tool=<name>=/path/to/bin/exe, remove the exe from path:
        path, exe = os.path.split(path_arg)
    elif path_arg and os.path.isdir(path_arg):
        # Someone passes us --tool=<name>=/path/to/bin/ (did not have exe)
        path, exe = path_arg, config_exe
    else:
        path, exe = '', ''

    if not path or not exe:
        util.error(
            f'Can not find path or exe for --tool={tool}: {name=} path={path_arg}'
        )
        return name

    path = os.path.abspath(path)
    if os.path.isdir(path):
        paths = os.environ.get('PATH', '').split(':')
        if path not in paths:
            util.info(f'--tool={tool} has path information, prepending PATH with: {path}')
            os.environ['PATH'] = path + ':' + os.environ.get('PATH', '')
        else:
            util.info(f'--tool={tool} has path information, but {path} already in $PATH')

    user_exe = os.path.join(path, exe)
    if not os.access(user_exe, os.X_OK):
        util.error(f'--tool setting for {tool}: {user_exe} is not an executable')
        return name

    user_exe = safe_shutil_which(user_exe)

    if update_config:
        if isinstance(config_exe, list):
            config['tools'][name]['exe'][0] = user_exe
            for index,value in enumerate(config_exe[1:]):
                # update all entries, if we can, if the value is also in 'path'
                # from our set --tool=Name=path/exe
                new_value = os.path.join(path, os.path.split(value)[1])
                if os.path.exists(new_value) and safe_shutil_which(new_value) and \
                   os.access(new_value, os.X_OK):
                    config['tools'][name]['exe'][index] = new_value
        else:
            config['tools'][name]['exe'] = user_exe
        util.debug(f'For {tool=}, tools config entry updated')

    util.debug(f'For {tool=}, final {user_exe=}')

    return name



'''opencos.commands.lec - Base class command handler for: eda lec ...

Intended to be overriden by Tool based classes (such as CommandLecYosys, etc)
'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

import os

from opencos import eda_extract_targets
from opencos.eda_base import CommandDesign, Tool


class CommandLec(CommandDesign):
    '''Base class command handler for: eda lec ...'''

    CHECK_REQUIRES = [Tool]
    error_on_no_files_or_targets = False
    error_on_missing_top = False # we'll override it.

    command_name = 'lec'

    def __init__(self, config: dict):
        CommandDesign.__init__(self, config=config)
        self.args.update({
            'designs': [],
            'synth': True,
            'flatten-all': True,
        })
        self.args_help.update({
            'designs': (
                'Set the two LEC comparison designs: --designs=<target1> --designs=<target2>,'
                ' use this arg twice'
            ),
            'synth': 'run synthesis on the two designs prior to running LEC',
            'flatten-all': (
                'arg passed to "synth" if run with --synth, to disable use --no-flatten-all'
            ),
        })

        self.synth_design_verilog_fpaths = ['', '']


    def do_it(self) -> None:
        '''Common do_it() method that child classes can use prior to customization'''

        # set_tool_defines() is from class Tool. Since that is not inherited yet, but
        # should be by any handlers, check on the existence of set_tool_defines, and
        # error if not present.
        if not all(isinstance(self, x) for x in self.CHECK_REQUIRES):
            self.error('CommandLec.do_it() requires a Tool to be in parent classes, but none is.',
                       f'{self.CHECK_REQUIRES=}')
            return

        # add defines for this job from Tool class if present
        self.command_safe_set_tool_defines() # (Command.command_safe_set_tool_defines)

        # dump our config to work-dir for debug
        self.write_eda_config_and_args()

        # Note - we do not support --export in LEC.
        # Derived classes can do the rest, can call CommandLec.do_it(self) as a first step.


    def process_tokens(self, tokens: list, process_all: bool = True,
                       pwd: str = os.getcwd()) -> list:
        unparsed = CommandDesign.process_tokens(
            self, tokens=tokens, process_all=process_all, pwd=pwd
        )

        if self.stop_process_tokens_before_do_it():
            return unparsed

        # we require there to be two --designs set.
        if not self.args['designs'] and len(self.args['designs']) != 2:
            self.error('Requires two designs via --designs=<target1> --designs=<target2>',
                       f'designs={self.args["designs"]}')
            return []


        # Before we do anything else, make sure the two designs actually exist.
        for design in self.args['designs']:
            # maybe it's a file?
            if os.path.isfile(design):
                pass
            elif design in eda_extract_targets.get_targets(partial_paths=[design], base_path=pwd):
                pass
            else:
                self.error(f'--designs={design}, value is not a file or target')

        if not self.args['top']:
            # Correct the 'top' name so it's eda.<target1>.<target2>.lec (shortnames)
            _, short_target1 = os.path.split(self.args['designs'][0])
            _, short_target2 = os.path.split(self.args['designs'][1])
            self.args['top'] = f'eda.{short_target1}.{short_target2}'

        # create our work dir
        self.create_work_dir()
        self.run_dep_commands()
        self.do_it()
        self.run_post_tool_dep_commands()
        return unparsed


    def get_synth_result_fpath(self, target: str) -> str:
        '''Derived classes must define. Given a synth target for one of the two

        designs to compare, return the location of the synthesis file'''
        raise NotImplementedError



''' opencos.tools.modelsim_ase - Used by opencos.eda for sim/elab commands w/ --tool=modelsim_ase.

Contains classes for CommandSimModelsimAse, CommandElabModelsimAse.

Note that this is for 32-bit Modelsim Student Edition. Consider using --tool=questa_fse instead.
'''

# pylint: disable=R0801 # (duplicate code in derived classes, such as if-condition return.)

import os

from opencos.tools.questa_common import CommonSimQuesta, CommonFListQuesta, \
    CommonWavesQuesta

class CommandWavesModelsimAse(CommonWavesQuesta):
    '''Handler for: eda waves --tool=modelsim_ase'''
    _TOOL = 'modelsim_ase'
    _EXE = 'vsim'

    def __init__(self, config: dict):
        CommonWavesQuesta.__init__(self, config=config)
        # repairs: override self._TOOL, and run get_versions() again.
        self._TOOL = 'modelsim_ase'


class CommandSimModelsimAse(CommonSimQuesta):
    '''CommandSimModelsimAse is a command handler for: eda sim --tool=modelsim_ase'''

    _TOOL = 'modelsim_ase'
    _EXE = 'vsim'
    use_vopt = False

    def __init__(self, config: dict):
        CommonSimQuesta.__init__(self, config=config)

        # repairs: override self._TOOL, and run get_versions() again.
        self._TOOL = 'modelsim_ase'

        self.shell_command = os.path.join(self.sim_exe_base_path, 'vsim')
        self.starter_edition = True
        self.args.update({
            'tool': self._TOOL, # override
            'gui': False,
        })


        self.args_help.update({
            'vopt': (
                'Boolean to enable/disable use of vopt step prior to vsim step'
                ' Note that vopt args can be controlled with --elab-args=<value1>'
                ' --elab-args=<value2> ...'
            )
        })


class CommandElabModelsimAse(CommandSimModelsimAse):
    '''CommandElabModelsimAse is a command handler for: eda elab --tool=modelsim_ase'''

    command_name = 'elab'

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-elaborate'] = True


class CommandLintModelsimAse(CommandSimModelsimAse):
    '''CommandLintModelsimAse is a command handler for: eda lint --tool=modelsim_ase'''

    command_name = 'lint'

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-compile'] = True
        self.args['stop-after-elaborate'] = True


class CommandFListModelsimAse(CommonFListQuesta):
    '''CommandFListModelsimAse is a command handler for: eda flist --tool=modelsim_ase'''

    def __init__(self, config: dict):
        CommonFListQuesta.__init__(self, config=config)
        self._TOOL = 'questa_fse'



'''opencos.util -- support global logging, argparser for printing (colors)'''

# pylint: disable=too-many-lines

import argparse
import atexit
import datetime
import json
import multiprocessing
import os
import re
import shutil
import subprocess
import sys
import time
import traceback

from enum import Enum
from pathlib import Path, PosixPath
from importlib import import_module
from dotenv import load_dotenv
from supports_color import supportsColor

import opencos
from opencos.files import safe_shutil_which
from opencos.utils import status_constants
from opencos.utils.str_helpers import strip_ansi_color

INITIAL_CWD = os.getcwd()

global_exit_allowed = False # pylint: disable=invalid-name
progname = "UNKNOWN"        # pylint: disable=invalid-name
progname_in_message = True  # pylint: disable=invalid-name
debug_level = 0             # pylint: disable=invalid-name
dot_f_files_expanded = []   # pylint: disable=invalid-name
env_files_loaded = []       # pylint: disable=invalid-name

args = {                    # pylint: disable=invalid-name
    'color' : bool(supportsColor.stdout),
    'emoji' : bool(getattr(supportsColor.stdout, 'level', 0) >= 2),
    'quiet' : False,
    'verbose' : False,
    'debug' : False,
    'fancy' : sys.stdout.isatty(),
    'warnings' : 0,
    'errors' : 0,
    'artifacts-json': True,
}
modified_args = {}

max_error_code = 0  # pylint: disable=invalid-name


class Colors:
    '''Namespace class for color printing help

    Avoid calling these directly, other than perhapas calling info(*txt, color=Colors.red)
    with 'color' set. It is preferred for outside callers to use one of the print_<color>(..)
    functions, or one of info|warning|error|debug
    '''

    off    = "\033[0m"

    # styling
    bold =      "\033[1m"
    faint =     "\033[2m"
    italic =    "\033[3m"
    underline = "\033[4m"
    blink =     "\033[5m"
    negative =  "\033[7m"
    crossed =   "\033[9m"

    # base colors, does not clear styling
    black =  "\033[30m"
    red =    "\033[31m"
    green =  "\033[32m"
    yellow = "\033[33m" # This looks orange/brown, but it's techincally yellow
    blue =   "\033[34m"
    purple = "\033[35m"
    cyan =   "\033[36m"
    white =  "\033[37m"
    light_gray = "\033[37m"
    foreground = "\033[39m"

    # bolds, shortname
    bred =      "\033[1;31m"
    bgreen =    "\033[1;32m"
    byellow =   "\033[1;33m"
    bblue =     "\033[1;34m"
    bpurple =   "\033[1;35m"
    bcyan =     "\033[1;36m"
    bwhite =    "\033[1;37m"
    bforeground = "\033[1;39m"

    # off/normal/clear + shortnames
    off_black =  "\033[0;30m"
    off_red =    "\033[0;31m"
    off_green =  "\033[0;32m"
    off_yellow = "\033[0;33m"
    off_blue =   "\033[0;34m"
    off_purple = "\033[0;35m"
    off_cyan =   "\033[0;36m"
    off_white =  "\033[0;37m"
    off_light_gray = "\033[0;37m"
    off_foreground = "\033[0;39m"

    # off/normal/clear + bold + shortnames
    off_bred =      "\033[0;1;31m"
    off_bgreen =    "\033[0;1;32m"
    off_byellow =   "\033[0;1;33m"
    off_bblue =     "\033[0;1;34m"
    off_bpurple =   "\033[0;1;35m"
    off_bcyan =     "\033[0;1;36m"
    off_bwhite =    "\033[0;1;37m"
    off_bforeground = "\033[0;1;39m"

    info = "\033[0;2;39m" # off + faint + foreground
    binfo = "\033[1;2;39m" # bold + faint + foreground
    iinfo = "\033[2;3;39m" # faint + italic + foreground
    biinfo = "\033[1;2;3;39m" # bold + faint + italic + foreground
    sinfo = "\033[2;39m" # faint + foreground (no OFF, leave prev style)

    err = "\033[0;31m" # off + red
    berr = "\033[1;31m" # bold + red

    warn = "\033[0;33m" # off + yellow
    bwarn = "\033[1;33m" # bold + yellow

    @staticmethod
    def color_text(text: str, color: str) -> str:
        '''Wraps 'text' (str) with color (one of red|green|yellow|foreground) prefix and

        color (normal) suffix. Disables color prefix/suffix wrapping if args['color']=False
        '''
        if args['color']:
            return color + text + "\033[0m" # (normal)
        return text

    def disable(self) -> None:
        '''Clears all color str in class Colors, to prevent print() methods that use
        util.Colors from printing color'''
        for x in dir(self):
            if not callable(getattr(self, x, None)) and isinstance(x, str) and \
               not x.startswith('_'):
                setattr(self, x, '')


def safe_emoji(emoji: str, default: str = '') -> str:
    '''Returns emoji character if args['emoji'] is True'''
    if args['emoji']:
        return emoji
    return default


class ArtifactTypes(Enum):
    '''Types that are allow-listed for artifacts.add* methods. If you don't use one of
    these, you'll get a warning'''
    TEXT = 0
    JSON = 1
    YAML = 2
    WAVEFORM = 3
    DOTF = 4
    TCL = 5
    SHELL = 6
    BITSTREAM = 7
    LICENSE = 8
    HTML = 9
    COVERAGE_DATABASE = 10
    UNKNOWN = 255

class Artifacts:
    '''Class to hold file artifacts, for logs generated by EDA, or other artifcats created
    by a specific tool.

    If enabled, will write out an artifacts.json file upon util.exit()
    '''

    ARTIFACTS_NAME = 'artifacts.json'
    EDA_ERRORS_NAME = 'eda_errors.json'

    data: dict = {}
    unresolved_data: dict = {}
    enabled: bool  = True
    artifacts_json_filepath: str = os.path.join('eda.work', ARTIFACTS_NAME)
    eda_errors_json_filepath: str = os.path.join('eda.work', EDA_ERRORS_NAME)
    tool_errors: dict = {}
    tool_warnings: dict = {}
    job_errors_or_warnings: dict = {}

    def _get_type_str(self, typ: str) -> (bool, str):
        '''Gets the name from ArtifactTypes, and returns tuple of (True, str)

        if the ArtifactTypes exists, else (False, str.upper())'''
        ret = getattr(ArtifactTypes, typ.upper(), '')
        if ret:
            return True, ret.name.lower()
        return False, typ.lower()

    def incr_tool_error(self, tool: str, count: int = 1) -> None:
        '''Increments self.tool_errors[tool]'''
        if tool not in self.tool_errors:
            self.tool_errors[tool] = count
        else:
            self.tool_errors[tool] += count

    def incr_tool_warning(self, tool: str, count: int = 1) -> None:
        '''Increments self.tool_warnings[tool]'''
        if tool not in self.tool_warnings:
            self.tool_warnings[tool] = count
        else:
            self.tool_warnings[tool] += count

    def add(self, name: str, typ: str, description: str = '') -> None:
        '''Adds file to artifacts'''

        if not self.enabled:
            return # artifacts dumping disabled.

        _, shortname = os.path.split(name)
        if shortname in self.data:
            return # artifact already registered

        type_exists, type_str = self._get_type_str(typ)
        if not type_exists:
            warning(f'Unknown artifacts {typ=} add for {name=} {description=}')

        self.data[shortname] = {
            'name': name,
            'type': type_str,
            'size': 0, # will be filled in when JSON written upon exit.
            'description': description,
        }

    def add_extension(
            self, search_paths: list, file_extension: str, typ: str, description: str = ''
    ) -> None:
        '''Adds a placeholder for file extensions to add to artifacts

        For example, when saving a Simulation waves, we may know it's a .fst or .vcd file,
        but we do not know what the testbench SV file has set in $dumpfile(filepath)
        '''
        type_exists, type_str = self._get_type_str(typ)
        if not type_exists:
            warning(f'Unknown artifacts {typ=} add_extension for {search_paths=}',
                    f'{file_extension=} {description=}')
        else:
            info(f'artifacts: add_extension "{file_extension}", search paths: {search_paths}')

        if isinstance(search_paths, (str, PosixPath)):
            sp = [search_paths]
        else:
            sp = search_paths
        self.unresolved_data[file_extension] = {
            'search_paths': sp,
            'type': type_str,
            'description': description
        }

    def set_artifacts_json_dir(self, dirpath: str) -> None:
        '''Sets the artifacts_json_filepath'''
        self.artifacts_json_filepath = os.path.join(dirpath, self.ARTIFACTS_NAME)
        self.eda_errors_json_filepath = os.path.join(dirpath, self.EDA_ERRORS_NAME)

    def reset(self, enable: bool = True) -> None:
        '''Clears internal data (dict). Called by util.process_tokens()'''
        self.data = {}
        self.unresolved_data = {}
        self.tool_errors = {}
        self.tool_warnings = {}
        self.job_errors_or_warnings = {}
        self.enabled = enable or args['artifacts-json']
        self.artifacts_json_filepath = os.path.join('eda.work', self.ARTIFACTS_NAME)
        self.eda_errors_json_filepath = os.path.join('eda.work', self.EDA_ERRORS_NAME)

    def _resolve_unresolved_data(self, ext: str) -> None:
        '''Find files for this extension that were registered using add_extension(...)

        adds to self.data'''
        entry = self.unresolved_data[ext]
        if not entry:
            return

        for search_path in entry['search_paths']:
            p = Path(search_path)
            # Don't add the JSON if it's artifacts.json or eda_errors.json:
            if p.resolve() == Path(self.artifacts_json_filepath).resolve():
                continue
            if p.resolve() == Path(self.eda_errors_json_filepath).resolve():
                continue
            for posix_filename in p.glob(f'*.{ext}'):
                self.add(
                    name=str(posix_filename), typ=entry['type'], description=entry['description']
                )

    def add_eda_errors_json(self) -> None:
        '''Adds eda_errors.json to artifacts'''

        safe_mkdir_for_file(self.eda_errors_json_filepath)

        # Save error/warning information to same directory:
        # Note - the global artifacts.tool_errors and tool_warnings are incremented by
        # Command.update_tool_warn_err_counts_from_log_lines(..)
        error_data = {
            'errors': args['errors'],
            'warnings': args['warnings'],
            'tool_errors': self.tool_errors,
            'tool_warnings': self.tool_warnings
        }
        if self.job_errors_or_warnings:
            error_data['job_errors_or_warnings'] = self.job_errors_or_warnings

        try:
            with open(self.eda_errors_json_filepath, 'w', encoding='utf-8') as f:
                json.dump(error_data, f, indent=4)
        except PermissionError:
            error(f'PermissionError trying to write file: {self.eda_errors_json_filepath}')
        except IOError as e:
            error(f'IOError while trying to write file: {self.eda_errors_json_filepath}, {e}')

        self.add(
            name=self.eda_errors_json_filepath, typ='json',
            description='opencos-eda EDA errors/warnings, and tool reported errors/warnings'
        )


    def write_json(self) -> None: # pylint: disable=too-many-branches
        '''Write out the artifacts.json file, called by util.exit()'''

        if not self.enabled:
            return # artifacts dumping disabled.

        if self.unresolved_data:
            for ext in self.unresolved_data:
                self._resolve_unresolved_data(ext)

        self.add_eda_errors_json()

        if not self.data:
            return # do nothing if we have no artifacts

        safe_mkdir_for_file(self.artifacts_json_filepath)

        # Update all file sizes:
        remove_keys = set()
        for key, entry in self.data.items():
            if entry['type'] == 'license':
                # License entries are metadata, not files - keep them as-is
                entry['size'] = 0
            elif os.path.isfile(entry['name']):
                entry['size'] = os.path.getsize(entry['name'])
            else:
                # file doesn't exist, remove it from artifacts.
                warning(f'Removing {key} ({entry["name"]}) from artifacts (file does not exist)')
                remove_keys.add(key)
        for key in remove_keys:
            del self.data[key]

        try:
            with open(self.artifacts_json_filepath, 'w', encoding='utf-8') as f:
                json.dump(self.data, f, indent=4)
        except PermissionError:
            error(f'PermissionError trying to write file: {self.artifacts_json_filepath}')
        except IOError as e:
            error(f'IOError while trying to write file: {self.artifacts_json_filepath}, {e}')

        info(f'Wrote artifacts JSON: {self.artifacts_json_filepath}')
        for entry in self.data.values():
            if entry["name"] and entry["type"] and \
               (args['errors'] or entry["name"] != self.eda_errors_json_filepath):
                # If the file is eda_errors_json_filepath, only info that line if
                # we had errors (we always save that file though, it's in artifacts)
                info(f'  --> {entry["name"]}  ({entry["type"]})')


artifacts = Artifacts()


class UtilLogger:
    '''Class for the util.global_log'''
    file = None
    filepath = ''
    time_last = 0 #timestamp via time.time()

    # disabled by default, eda.py enables it. Can also be disabled via
    # util's argparser: --no-default-log, --logfile=<name>, or --force-logfile=<name>
    default_log_enabled = False
    default_log_filepath = os.path.join('eda.work', 'eda.log')
    default_log_disable_with_args = ['-h', '--help'] # common argparse help
    enable = True

    def clear(self) -> None:
        '''Resets some internals for logging, but preserve others'''
        self.file = None
        self.filepath = ''
        self.time_last = 0

    def stop(self) -> None:
        '''Closes open log, resets some internals'''
        if self.file:
            self.write_timestamp(f'stop - {self.filepath}')
            info(f"Closing logfile: {self.filepath}")
            self.file.close()
        self.clear()

    def start(self, filename: str, force: bool = False) -> None:
        '''Starts (opens) log'''
        if not self.enable:
            self.file = None
            return
        if not filename:
            error('Trying to start a logfile, but filename is missing')
            return
        if os.path.exists(filename):
            if force:
                debug(f"Overwriting logfile '{filename}', which exists, due to --force-logfile.")
            else:
                error(f"The --logfile path '{filename}' exists.  Use --force-logfile",
                      "(vs --logfile) to override.")
                return
        else:
            safe_mkdir_for_file(filename)
        try:
            self.file = open(filename, 'w', encoding='utf-8') # pylint: disable=consider-using-with
            debug(f"Opened logfile '{filename}' for writing")
            self.filepath = filename
            self.write_timestamp(f'start - {self.filepath}')
            # add to global artifacts:
            artifacts.add(name=filename, typ='text', description='EDA stdout logfile')
        except PermissionError:
            error(f'PermissionError opening log: {filename}')
            self.clear()
        except IOError as e:
            error(f'IOError while opening log: {filename}, {e}')
            self.clear()
        except Exception as e:
            error(f"Error opening '{filename}' for writing, {e}")
            self.clear()

    def write_timestamp(self, text: str = "") -> None:
        '''Writes timestamp to opened log'''
        if not self.file:
            return
        dt = datetime.datetime.now().ctime()
        print(f"INFO: [{progname}] Time: {dt} {text}", file=self.file)
        self.time_last = time.time()

    def write(self, text: str, end: str = '\n') -> None:
        '''Writes text to opened log'''
        if not self.file:
            return

        if ((time.time() - self.time_last) > 10) and \
           any(text.startswith(x) for x in [
               f"DEBUG: [{progname}]",
               f"INFO: [{progname}]",
               f"WARNING: [{progname}]",
               f"ERROR: [{progname}]"]):
            self.write_timestamp()
        print(strip_ansi_color(text), end=end, file=self.file)
        self.file.flush()
        os.fsync(self.file)


global_log = UtilLogger()


def start_log(filename: str, force: bool = False) -> None:
    '''Starts the global_log, if not already started'''
    global_log.start(filename=filename, force=force)

def write_log(text: str, end: str = '\n') -> None:
    '''Writes to the global_log, if started'''
    global_log.write(text=text, end=end)

def stop_log() -> None:
    '''Stops/closed the global_log'''
    global_log.stop()


atexit.register(stop_log)


def get_thread_budget() -> int:
    '''
    Returns an int, a la $(nproc), but takes the sched_getaffinity if available (Linux) to see
    what the OS actually allows this process to use. Otherwise, use cpu_count.
    '''
    try:
        return max(1, len(os.sched_getaffinity(0)))
    except AttributeError:
        return max(1, multiprocessing.cpu_count())
    except Exception:
        return 1


def get_tool_args_threads(total_budget: int, internal_parallelism: int) -> int:
    '''
    Returns an int, used by CommandParallel tools, to avoid --parallel=2 (yes 2x!)
    but then each of those oversubscribes to what the system advertises.

    Minimum return value 1. Example use:
        child_threads = util.get_tool_args_threads(
            self.args.get('threads', 1) / self.args.get('parallel', 1))
        )
        # now call sub eda command with: f'-j {child_threads}'
    '''
    # Ensure at least 1 thread minimum
    return max(1, total_budget // internal_parallelism)


def get_argparse_bool_action_kwargs() -> dict:
    '''Returns args for BooleanOptionalAction kwargs for an ArgumentParser.add_argument

    This is mostly for python compatibility to support --some-enable and --no-some-enable
    '''
    bool_kwargs = {}
    x = getattr(argparse, 'BooleanOptionalAction', None)
    if x is not None:
        bool_kwargs['action'] = x
    else:
        bool_kwargs['action'] = 'store_true'
    return bool_kwargs


def get_argparser() -> argparse.ArgumentParser:
    '''Returns the opencos.util ArgumentParser'''
    parser = argparse.ArgumentParser(
        prog=f'{safe_emoji("🔎 ")}opencos common options', add_help=False, allow_abbrev=False
    )
    # We set allow_abbrev=False so --force-logfile won't try to attempt parsing shorter similarly
    # named args like --force, we want those to go to unparsed list.
    # For bools, support --color and --no-color with this action=argparse.BooleanOptionalAction
    # -- however Python3.8 and older does not support this, so as a workaround, use kwargs for
    #    boolean actions:
    bool_action_kwargs = get_argparse_bool_action_kwargs()

    parser.add_argument('--version', default=False, action='store_true',
                        help=('Shows our version:'
                              f' {opencos.__version__} ({opencos.__pyproject_name__})')
                        )
    parser.add_argument('--color', **bool_action_kwargs, default=bool(supportsColor.stdout),
                        help='Use shell colors for info/warning/error messaging')
    parser.add_argument('--emoji', **bool_action_kwargs, default=args['emoji'],
                        help=f'Support emojis in terminal{safe_emoji(" 💪")}')
    parser.add_argument('--quiet', **bool_action_kwargs, default=args['quiet'],
                        help='Do not display info messaging')
    parser.add_argument('--verbose', **bool_action_kwargs, default=args['verbose'],
                        help='Display additional messaging level 2 or higher')
    parser.add_argument('--fancy', **bool_action_kwargs, default=args['fancy'])
    parser.add_argument('--debug', **bool_action_kwargs, default=args['debug'],
                        help='Display additional debug messaging level 1 or higher')
    parser.add_argument('--debug-level', type=int, default=0,
                        help='Set debug level messaging (default: 0)')
    parser.add_argument('--logfile', type=str, default=None,
                        help=('Write eda messaging to safe logfile that will not be overwritten'
                              ' (default disabled)'))
    parser.add_argument('--force-logfile', type=str, default=None,
                        help='Set to force overwrite the logfile')
    parser.add_argument('--default-log', **bool_action_kwargs,
                        default=global_log.default_log_enabled,
                        help=('Enable/Disable default logging to'
                              f' {global_log.default_log_filepath}. Default logging is disabled'
                              ' if --logfile or --force-logfile is set'))
    parser.add_argument('--no-respawn', action='store_true',
                        help=('Legacy mode (default respawn disabled) for respawning eda.py'
                              ' using $OC_ROOT/bin'))
    parser.add_argument('--artifacts-json', **bool_action_kwargs,
                        help='Store a work-dir/artifacts.json file of known tool output files')
    parser.add_argument('-f', '--input-file', default=[], action='append',
                        help=(
                            'Input .f file to be expanded as eda'
                            ' args/defines/incdirs/files/targets'))
    parser.add_argument('--env-file', default=[], action='append',
                        help=(
                            "dotenv file(s) to pass ENV vars, (default: .env loaded first,"
                            " subsequent files' vars override .env"
                        ))
    return parser


def get_argparser_args_list(parser: object = None) -> list:
    '''Returns list of all args, all items will include the -- prefix (--help, etc)'''
    if not isinstance(parser, argparse.ArgumentParser):
        return []
    return list(vars(parser).get('_option_string_actions', {}).keys())


def get_argparsers_args_list(parsers: list) -> list:
    '''Returns list of all args from list of parsers, all items have -- prefix (--help, etc)'''
    _args_list = []
    for parser in parsers:
        _args_list.extend(get_argparser_args_list(parser))
    return list(dict.fromkeys(_args_list)) # uniqify, preserve order


def get_argparser_short_help(parser: object = None) -> str:
    '''Returns short help for our ArgumentParser'''
    if not parser:
        parser = get_argparser()

    if not args['color']:
        # Since python3.14 doesn't care about our custom color settings,
        # need to remove any ANSI colors from argparse help formatter:
        full_lines = strip_ansi_color(parser.format_help()).split('\n')
    else:
        full_lines = parser.format_help().split('\n')

    lineno = 0
    for lineno, line in enumerate(full_lines):
        # Again, strip any ANSI colors when searching for starting text:
        # - options:
        # - optional arguments:
        if args['color']:
            line = strip_ansi_color(line)
        if any(line.startswith(x) for x in ('options:', 'optional arguments:')):
            break

    # skip the line that says 'options:', replace with the progname:
    return f'{Colors.off_cyan}{parser.prog}:{Colors.off}\n' + '\n'.join(full_lines[lineno + 1:])


def process_token(arg: list) -> bool:
    '''Returns true if we parsed arg (list, because we may pass --arg value)

    This is legacy holdover for oc_cli.py, that would process one token at a time.
    Simply run through our full argparser.
    '''
    _, unparsed = process_tokens(arg)
    if not unparsed: # empty list
        debug(f"Processed token or pair: {arg}")
        return True
    return False


def load_env_file(env_file: str) -> None:
    '''Handles .env file (from util CLI args --env-file)'''
    if os.path.isfile(env_file):
        load_dotenv(env_file, override=True)
        if os.path.abspath(env_file) not in env_files_loaded:
            env_files_loaded.append(os.path.abspath(env_file))
    else:
        warning(f'--env-file {env_file} does not exist and is not loaded.')


def patch_args_for_dir(tokens: list, patch_dir: str, caller_info: str) -> list:
    '''Given list of args, attempt to correct for relative dir'''

    # deal with relative path args or files.
    # Note the dot-f file could have been in a different directory (-f=path/to/my.f)
    # As a workaround to deal with relative paths, attempt to replace relative
    # path args within this dotf contents - so they are relative the dotf dir:
    if not os.path.isdir(patch_dir):
        return tokens

    ret = []
    for word in tokens:
        if word.startswith('-') and '=' in word:
            parts = word.split('=')
            leftarg = parts[0] + '='
            word = '='.join(parts[1:])
        elif word.startswith('+incdir+'):
            # do for +incdir+ too
            leftarg = '+incdir+'
            word = word[len('+incdir+'):]
        else:
            leftarg = ''

        if word and not os.path.isabs(word) and \
           os.path.exists(os.path.join(patch_dir, word)):
            # fix relative path of word, or --arg=word
            word = os.path.abspath(os.path.join(patch_dir, word))
            info(f'Using relative path {patch_dir} for arg/token: {leftarg}{word}',
                 f'{caller_info}')
        ret.append(f'{leftarg}{word}')
    return ret

def read_tokens_from_dot_f(filepath: str, caller_info: str = '', verbose: bool = False) -> list:
    '''Returns list of tokens from a .f file, with ENV vars expanded'''

    # Let's defer 'info' printing out what input files were opened until after
    # args['quiet'] and debug is resolved (which may be in these .f files)
    start_str = f"Opening -f / --input-file '{filepath}' for contents {caller_info}"
    if verbose:
        info(start_str)
    else:
        debug(start_str)
    if not os.path.isfile(filepath):
        error(f'-f (or --input-file): {filepath} does not exist {caller_info}',
              error_code=status_constants.EDA_GENERAL_FILE_NOT_FOUND)
        return []
    if os.path.abspath(filepath) in dot_f_files_expanded:
        error(f'-f (or --input-file): {filepath} has already been expanded',
              f'cannot traverse again (duplicate arg or nested .f files) {caller_info}')
    else:
        dot_f_files_expanded.append(os.path.abspath(filepath))
    tokens = []
    dotf_file_dir, _ = os.path.split(filepath)
    with open(filepath, encoding='utf-8') as f:
        for line in f:
            line = os.path.expandvars(line.strip())
            if not line or line.startswith('#') or line.startswith('//'):
                continue
            words = []
            if line.startswith("'") and line.endswith("'"):
                # treat as single word:
                words = [line]
            else:
                words = line.split()
            tokens.extend(patch_args_for_dir(
                tokens=words, patch_dir=dotf_file_dir, caller_info=f"(from dotf {filepath})"
            ))
    return tokens


def process_debug_args(parsed: argparse.Namespace) -> None:
    '''Sets global debug_level based on parsed args'''
    global debug_level # pylint: disable=global-statement

    if parsed.debug_level:
        set_debug_level(parsed.debug_level)
    elif parsed.debug:
        set_debug_level(1)
    elif parsed.debug is False:
        # not None, explicitly set to False via --no-debug
        debug_level = 0


def process_tokens( # pylint: disable=too-many-branches,too-many-statements
        tokens: list, caller_info: str = ''
) -> (argparse.Namespace, list):
    '''Processes tokens (unparsed args list) on util's ArgumentParser

    Returns tuple of (parsed Namespace, unparsed args list)
    '''
    global debug_level # pylint: disable=global-statement
    debug_level = 0

    # Deal with --version, --debug, --debug-level, --env-file, -f/--input-file(s)
    # tokens first, for now put dot-f file contents in front of of tokens, do this first
    # in a separate custom argparser:
    bool_action_kwargs = get_argparse_bool_action_kwargs()
    parser = argparse.ArgumentParser(
        prog='opencos -f/--input-file', add_help=False, allow_abbrev=False
    )
    parser.add_argument('--version', default=False, action='store_true')
    parser.add_argument('--debug', **bool_action_kwargs,
                        help='Display additional debug messaging level 1 or higher')
    parser.add_argument('--debug-level', type=int, default=0,
                        help='Set debug level messaging (default: 0)')
    parser.add_argument('--env-file', default=[], action='append',
                        help=(
                            "dotenv file(s) to pass ENV vars, (default: .env loaded first,"
                            " subsequent files' vars override .env"
                        ))
    parser.add_argument('-f', '--input-file', default=[], action='append',
                        help=(
                            'Input .f file to be expanded as eda args, defines, incdirs,'
                            ' files, or targets.'
                        ))
    try:
        parsed, unparsed = parser.parse_known_args(tokens + [''])
        tokens2 = list(filter(None, unparsed))
    except argparse.ArgumentError:
        error(f'util -f/--input-file, problem attempting to parse_known_args for: {tokens}')

    if parsed.version:
        # Stop processing, return nothing, caller can print version and exit.
        return parsed, []

    process_debug_args(parsed=parsed)
    debug(f'util.process_tokens: {parsed=} {unparsed=} from: {tokens}')

    if os.path.isfile(str(Path('.env'))):
        parsed.env_file.insert(0, '.env')
    if parsed.env_file:
        for env_file in parsed.env_file:
            load_env_file(env_file)

    if parsed.input_file:
        dotf_tokens = []
        for filepath in parsed.input_file:
            dotf_tokens.extend(read_tokens_from_dot_f(
                filepath=filepath, caller_info=caller_info
            ))

        # put the .f files before the unparsed args.
        tokens2 = dotf_tokens + tokens2

        # recurse until we've resolved nested .f files.
        return process_tokens(tokens=tokens2, caller_info=f'(from {parsed.input_file[-1]})')


    # Continue with all normal parsing beyond --debug and -f/--input-file,
    # Note that we re-parse everything in case there was --debug or --debug-level in
    # the .f file(s), or anything else that this argparser would pick up.
    parser = get_argparser()
    try:
        parsed, unparsed = parser.parse_known_args(tokens2 + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        error(f'util: problem attempting to parse_known_args for tokens={tokens2}')

    process_debug_args(parsed=parsed)

    if parsed.input_file:
        warning(f'python error, nested -f/--input-file(s) {parsed.input_file} should',
                'have already been resolved')

    if not parsed.color:
        Colors.disable(Colors) # strip strings in Colors class

    # clear existing artifacts dicts (mostly for pytests repeatedly calling eda.main),
    # set artifacts.enabled based on args['artifacts-json']
    artifacts.reset(enable=parsed.artifacts_json)

    if parsed.force_logfile:
        start_log(parsed.force_logfile, force=True)
    elif parsed.logfile:
        start_log(parsed.logfile, force=False)
    elif parsed.default_log and \
         not parsed.version and \
         not any(x in unparsed for x in global_log.default_log_disable_with_args) and \
         (parsed.force_logfile is None and parsed.logfile is None):
        # Use a forced logfile in the eda.work/eda.log:
        # avoid this if someone has --help arg not yet parsed.
        # avoid this if someone called --version, b/c that will print and exit.
        start_log(global_log.default_log_filepath, force=True)

    parsed_as_dict = vars(parsed)
    for key, value in parsed_as_dict.items():
        key = key.replace('_', '-')
        if key in args and value is not None:
            if value != args.get(key, None):
                modified_args[key] = True
            args[key] = value # only update with non-None values to our global 'args' dict

    for filepath in env_files_loaded:
        info(f"Loaded --env-file '{os.path.relpath(filepath)}'")
    for filepath in dot_f_files_expanded:
        info(f"Processed -f / --input-file '{os.path.relpath(filepath)}' for contents")

    return parsed, unparsed

# ********************
# fancy support
# In fancy mode, we take the bottom _FANCY_LINES lines of the screen to be written using
# fancy_print, while the lines above that show regular scrolling content (via info, debug, warning,
# error above). User should not use print() when in fancy mode

_FANCY_LINES = []

def fancy_start(fancy_lines: int = 4, min_vanilla_lines: int = 4) -> None:
    '''Starts fancy line support. This is not called by util internally

    It is called by an opencos.eda Command handling class, it can check if util.args['fancy']
    is set
    '''
    _, lines = shutil.get_terminal_size()
    if fancy_lines < 2:
        error("Fancy mode requires at least 2 fancy lines")
    if fancy_lines > (lines - min_vanilla_lines):
        error(f"Fancy mode supports at most {(lines - min_vanilla_lines)} fancy lines, given",
              f"{min_vanilla_lines} non-fancy lines")
    if _FANCY_LINES:
        error("We are already in fancy line mode, cannot call fancy_start() again")
    for _ in range(fancy_lines - 1):
        print() # create the requisite number of blank lines
        _FANCY_LINES.append("")
    print("", end="") # the last line has no "\n" because we don't want ANOTHER blank line below
    _FANCY_LINES.append("")
    # the cursor remains at the leftmost character of the bottom line of the screen

def fancy_stop() -> None:
    '''Stops fancy mode. Intended to be called by an opencos.eda Command handling class'''
    if not _FANCY_LINES:
        # don't do anything if we aren't in fancy mode.
        return

    # user is expected to have painted something into the fancy lines, we can't "pull down"
    # the regular lines above, and we don't want _FANCY_LINES blank or garbage lines either,
    # that's not pretty
    _FANCY_LINES.clear()
    # since cursor is always left at the leftmost character of the bottom line of the screen,
    # which was one of the fancy lines which now has the above-mentioned "something", we want
    # to move one lower
    print()


def fancy_print(text: str, line: int) -> None:
    '''Fancy print, intended to be called by an opencos.eda Command handling class'''

    # strip any newline, we don't want to print that
    text = text.rstrip("\n")
    lines_above = len(_FANCY_LINES) - line - 1
    if lines_above:
        print(
            (f"\033[{lines_above}A" # move cursor up
             f"{text}\033[1G" # desired text, then move cursor to the first character of the line
             f"\033[{lines_above}B" # move the cursor down
             ),
            end="", flush=True
        )
    else:
        print(
            f"{text}\033[1G", # desired text, then move cursor to the first character of the line
            end="", flush=True
        )
    _FANCY_LINES[line] = text

def print_pre() -> None:
    '''called by all util info/warning/debug/error. Handles fancy mode'''
    # stuff we do before printing any line
    if not _FANCY_LINES:
        return

    # Also, note that in fancy mode we don't allow the "above lines" to be partially written, they
    # are assumed to be full lines ending in "\n"
    # As always, we expect the cursor was left in the leftmost character of bottom line of screen
    print(
        (f"\033[{len(_FANCY_LINES)-1}A" # move the cursor up to where the first fancy line is drawn
         f"\033[0K"                     # clear the old fancy line 0
         ),
        end="", flush=True
    )

def print_post(text: str, end: str) -> None:
    '''called by all util info/warning/debug/error. Handles fancy mode'''
    # stuff we do after printing any line
    if _FANCY_LINES:
        # we just printed a line, including a new line, on top of where fancy line 0 used to be,
        # so cursor is now at the start of fancy line 1. move cursor down to the beginning of the
        # final fancy line (i.e. standard fancy cursor resting place)
        for x, line in enumerate(_FANCY_LINES):
            print("\033[0K", end="") # erase the line to the right
            print(line, flush=True,
                  end=('' if x == (len(_FANCY_LINES) - 1) else '\n'))

        print("\033[1G", end="", flush=True)
    if global_log.file:
        write_log(text, end=end)

def print_wrap(text: str, end: str = '\n') -> None:
    '''Same a print_color, but without a color change'''
    print_pre()
    print(text + Colors.off, end=end, flush=True)
    print_post(text, end)

def print_color(text: str, color: str, end: str = '\n') -> None:
    '''Note that color(str) must be one of Colors.[red|green|yellow|normal]'''
    print_pre()
    print(Colors.color_text(text, color), end=end, flush=True)
    print_post(text, end)

def print_foreground_color(text: str, end: str = '\n') -> None:
    '''Print text as foreground color, goes back to normal color'''
    print_pre()
    print(Colors.color_text(text, color=Colors.foreground), end=end, flush=True)
    print_post(text, end)


def set_debug_level(level) -> None:
    '''Sets global debug level, sets args['debug'] and args['verbose']'''
    global debug_level # pylint: disable=global-statement
    debug_level = level
    args['debug'] = level > 0
    args['verbose'] = level > 1
    info(f"Set debug level to {debug_level}")


def debug(
        *text, level: int = 1, start: object = None, end: str = '\n', color=Colors.foreground
) -> None:
    '''Print debug messaging (in foreground color if possible). If args['debug'] is false,
    prints nothing.

    *text: (positional str args) to be printed
    level: (int) debug level to decide if printed or not.
    start: (optional str) prefix to message; if None: chooses default start str
    end: (optional str) suffix to print

    Note these messages append to global logging (but require args['debug'] to be set)
    '''
    if start is None:
        start = "DEBUG: " + (f"[{progname}] " if progname_in_message else "")
    if args['debug'] and \
       (((level==1) and args['verbose']) or (debug_level >= level)):
        print_color(f"{start}{' '.join(list(text))}", color=color, end=end)


def info(*text, start: object = None, end='\n', color=None) -> None:
    '''Print information messaging (in green if possible). If args['quiet'], prints nothing.

    *text: (positional str args) to be printed
    start: (optional str) prefix to message; if None: chooses default start str
    end: (optional str) suffix to print

    Note these messages append to global logging even if args['quiet'] is set
    '''
    if color is None:
        color = Colors.info
    if start is None:
        start = (
            f"{Colors.off}{Colors.green}INFO: " + \
            (f"[{progname}] " if progname_in_message else "")
        )
    if not args['quiet']:
        print_wrap(f"{start}{color}{' '.join(list(text))}", end=end)

def warning(*text, start: object = None, end: str = '\n') -> None:
    '''Print warning messaging (in yellow if possible).

    *text: (positional str args) to be printed
    start: (optional str) prefix to message; if None: chooses default start str
    end: (optional str) suffix to print

    Note these messages append to global logging. Increments global args['warnings'] int.
    '''
    if start is None:
        start = (
            f"{Colors.off}{Colors.bwarn}WARNING: " + \
            (f"[{progname}] " if progname_in_message else "")
        )
    args['warnings'] += 1
    print_wrap(f"{start}{Colors.warn}{' '.join(list(text))}", end=end)


def error(
        *text,
        error_code: int = status_constants.EDA_DEFAULT_ERROR,
        do_exit: bool = True, start: object = None, end: str = '\n'
) -> int:
    '''Print error messaging (in red if possible).

    *text: (positional str args) to be printed
    error_code: (int) shell style return code (non-zero is error, but prefer > 1 b/c those are
                python exceptions)
    do_exit: (bool) if True will call exit based on global_exit_allowed.
    start: (optional str) prefix to message; if None: chooses default start str
    end: (optional str) suffix to print

    Note these messages append to global logging. Increments global args['errors'] int.
    '''
    global max_error_code # pylint: disable=global-statement

    if start is None:
        start = f"{Colors.off}{Colors.berr}ERROR: "
        if progname_in_message:
            start += f"[{progname}] "
        start += safe_emoji("❌ ")
    args['errors'] += 1
    max_error_code = max(max_error_code, error_code)
    print_wrap(f"{start}{Colors.err}{' '.join(list(text))}", end=end)
    if do_exit:
        if args['debug']:
            print(traceback.print_stack())
        # Call our overriden-builtin function for 'exit':
        return exit(error_code) # pylint: disable=consider-using-sys-exit

    if error_code is None:
        return 0
    return abs(int(error_code))


class AlreadyMessage:
    '''
    Namespace, so we only make the same warnings once, aka ToolVerilator may be called > 1 times
    but we only want to see those warnings once, vs N times for every N class handles created.

    Used by warn_once(warn_name, "text")
    '''
    info_names: set = set()
    warn_names: set = set()


def info_once(*text, info_name: str, **kwargs) -> None:
    '''Wrapper for util.info, only runs if info_name hasn't happened before'''
    if info_name not in AlreadyMessage.info_names:
        info(*text, **kwargs)
        AlreadyMessage.info_names.add(info_name)


def warn_once(*text, warn_name: str, **kwargs) -> None:
    '''Wrapper for util.warning, only runs if warn_name hasn't happened before'''
    if warn_name not in AlreadyMessage.warn_names:
        warning(*text, **kwargs)
        AlreadyMessage.warn_names.add(warn_name)


def get_return_code() -> int:
    '''Checks global max_error_code and args['errors']'''
    if args['errors']:
        if max_error_code == 0:
            return 255
        return max_error_code
    return 0


def exit( # pylint: disable=redefined-builtin
        error_code: int = 0, quiet: bool = False
) -> int:
    '''sys.exit(int) wrapper, returns the error_code if global_exit_allowed=False'''

    # Save out artifacts file:
    artifacts.write_json()

    if global_exit_allowed:
        if not quiet:
            info_color = Colors.info
            start = safe_emoji('🔚 ')
            if args['errors']:
                info_color = Colors.red
                start = safe_emoji('❗ ')
            elif args['warnings']:
                info_color = Colors.yellow
            info(
                f"{start}Exiting with {Colors.bold}{args['warnings']} warnings{info_color},",
                f"{Colors.bold}{args['errors']} errors",
                color=info_color
            )
        sys.exit(error_code)

    if error_code is None:
        return 0

    return abs(int(error_code))


def getcwd():
    '''Wrapper for os.getcwd() for current working dir'''
    try:
        return os.getcwd()
    except Exception as e:
        error(f"Unable to getcwd(), did it get deleted from under us? Exception: {e}")
        return None

_OC_ROOT = None       # pylint: disable=invalid-name
_OC_ROOT_SET = False  # pylint: disable=invalid-name

def get_oc_root(error_on_fail: bool = False) -> None:
    '''Returns a str or False for the root directory of *this* repo.

    If environment variable OC_ROOT is set, that is used instead, otherwise attempts to use
    `git rev-parse`
    '''

    global _OC_ROOT      # pylint: disable=global-statement
    global _OC_ROOT_SET  # pylint: disable=global-statement
    # if we've already run through here once, just return the memorized result
    if _OC_ROOT_SET:
        return _OC_ROOT

    # try looking for an env var
    s = os.environ.get('OC_ROOT')
    if s:
        debug(f'get_oc_root() -- returning from env: {s=}')
        _OC_ROOT = s.strip()
    else:
        # try asking GIT
        cp = subprocess.run(
            'git rev-parse --show-toplevel', stdout=subprocess.PIPE, stderr=subprocess.DEVNULL,
            shell=True, check=False, universal_newlines=True
        )
        if cp.returncode != 0:
            # TODO(drew): at some point, address the fact that not all repos are oc_root.
            # Is this function asking for the repo we are in?  or a pointer to the oc_root which
            # maybe elsewhere on the system?
            print_didnt_find_it = debug
            if error_on_fail:
                print_didnt_find_it = error
            print_didnt_find_it('Unable to get a OC_ROOT directory using git rev-parse')
        else:
            _OC_ROOT = cp.stdout.strip()
            if sys.platform.startswith('win'):
                _OC_ROOT = _OC_ROOT.replace('/', '\\') # git gives us /, but we need \

    # there is no sense running through this code more than once
    _OC_ROOT_SET = True
    return _OC_ROOT


def safe_cp(source: str, destination: str, create_dirs: bool = False) -> None:
    '''shutil.copy2 wrapper to optionally make the destination directories'''
    try:
        # Infer if destination is a directory
        if destination.endswith('/') or os.path.isdir(destination):
            if not os.path.exists(destination) and create_dirs:
                os.makedirs(destination, exist_ok=True)
            destination = os.path.join(destination, os.path.basename(source))
        else:
            # Ensure parent directory exists if needed
            parent_dir = os.path.dirname(destination)
            if create_dirs and parent_dir and not os.path.exists(parent_dir):
                os.makedirs(parent_dir, exist_ok=True)
        # actually copy the file
        shutil.copy2(source, destination)
        info(f"Copied {source} to {destination}")
    except Exception as e:
        print(f"Error copying file from '{source}' to '{destination}': {e}")


def safe_mkdir(path : str) -> None:
    '''Attempt to make dir at path, and make all subdirs up to that path'''
    if os.path.exists(path):
        return
    left, _ = os.path.split(os.path.relpath(path))
    if left and left not in ['.', '..', os.path.sep]:
        safe_mkdir(left)
    try:
        os.mkdir(path)
    except FileExistsError:
        pass
    except Exception as e1:
        try:
            os.system(f'mkdir -p {path}')
        except Exception as e2:
            error(f'unable to mkdir {path=}, exceptions {e1}, {e2=}')


def safe_mkdirs(base : str, new_dirs : list) -> None:
    '''Create new_dirs at base'''
    for p in new_dirs:
        safe_mkdir( os.path.join(base, p) )


def safe_mkdir_for_file(filepath: str) -> None:
    '''Given a new filepath, create dir for that filepath'''
    left, _ = os.path.split(filepath)
    if left:
        safe_mkdir(left)


def import_module_from_string(full_module_name: str) -> None:
    """
    Imports a pymodule given its full name as a str.

    Args:
        full_module_name: The full name of the pymodule,
                           e.g., "module.submodule".

    Returns:
        The imported module, or None if an error occurs.
    """
    try:
        return import_module(full_module_name)
    except Exception as e:
        error(f"Error importing pymodule {full_module_name=}: {e}")
        return None


def import_class_from_string(full_class_name: str) -> None:
    """
    Imports a class given its full name as a str.

    Args:
        full_class_name: The full name of the class,
                         e.g., "module.submodule.ClassName".

    Returns:
        The imported class, or None if an error occurs.
    """
    try:
        module_path, class_name = full_class_name.rsplit(".", 1)
        module = import_module(module_path)
        return getattr(module, class_name)
    except Exception as e:
        error(f"Error importing class {full_class_name=}: {e}")
        return None


class ShellCommandList(list):
    '''Wrapper around a list, of str that we'll run as a subprocess command

    included member vars for:
      - tee_path, to save a log from this subprocess commands list
      - work_dir - in case we want to run this from non-default location.
    '''
    def __init__(self, obj: object = None, tee_fpath: str = '', work_dir: str = ''):
        super().__init__(obj)
        for k in ('tee_fpath', 'work_dir'):
            setattr(self, k, getattr(obj, k, None))
        if tee_fpath:
            self.tee_fpath = tee_fpath
        if work_dir:
            self.work_dir = work_dir


def write_shell_command_file(
        dirpath : str, filename : str, command_lists : list, line_breaks : bool = False
) -> None:
    ''' Writes new file at {dirpath}/{filename} as a bash shell command, using command_lists
    (list of lists)

    -- dirpath (str)        -- directory where file is written (usually eda.work/{target}_sim
    -- filename (str)       -- filename, for example compile_only.sh
    -- command_lists (list) -- list of (list or ShellCommandList), each item in the list is a
                               list of commands (aka, how subprocess.run(args) uses a list of
                               commands.
    -- line_breaks (bool)   -- Set to True to have 1 word per line in the file followed by a line
                               break. Default False has an entry in command_lists all on a single
                               line.

    Returns None, writes the file and chmod's it to 0x755.
    '''

    # command_lists should be a list-of-lists.
    bash_path = safe_shutil_which('bash')
    assert isinstance(command_lists, list), f'{command_lists=}'
    fullpath = os.path.join(dirpath, filename)
    try:
        with open(fullpath, 'w', encoding='utf-8') as f:
            if not bash_path:
                bash_path = "/bin/bash" # we may not get far, but we'll try
            f.write('#!' + bash_path + '\n\n')
            for obj in command_lists:
                assert isinstance(obj, list), \
                    f'{obj=} (obj must be list/ShellCommandList) {command_lists=}'
                clist = list(obj).copy()
                tee_fpath = getattr(obj, 'tee_fpath', None)
                if tee_fpath:
                    # Note the | tee foo.log will ruin bash error codes, so if we're bash is
                    # available, we'll check that ${PIPESTATUS} is 0 to percolate the
                    # a non-zero on the first command (sim.exe).
                    if safe_shutil_which('bash'):
                        clist.append(f'2>&1 | tee {tee_fpath}' + ' && test ${PIPESTATUS} -eq 0')
                    else:
                        clist.append(f'2>&1 | tee {tee_fpath}')

                if clist:
                    if line_breaks:
                        # line_breaks=True - have 1 word per line, followed by \ and \n
                        sep = " \\" + "\n"
                        f.write(sep.join(clist))
                        f.write(" \n")
                    else:
                        # line_break=False (default) - all words on 1 line.
                        f.write(' '.join(clist))
                        f.write(" \n")
                else:
                    f.write("\n")
            f.write("\n")
            f.close()
            os.chmod(fullpath, 0o755)
    except PermissionError:
        error(f'PermissionError trying to write file: {fullpath}')
    except IOError as e:
        error(f'IOError while trying to write file: {fullpath}, {e}')


def get_inferred_top_module_name(module_guess: str, module_fpath: str) -> str:
    '''Returns the best guess as the 'top' module name name, given a fpath where

    module_fpath = /some/path/to/module_guess.[v|sv]

    Use the module_guess if it exists in the file as: module <module_guess>
    Otherwise use the last observed: module <best_guess>
    Otherwise use blank str
    '''

    best_guess = ''
    if not os.path.isfile(module_fpath):
        return ''
    with open(module_fpath, encoding='utf-8') as f:
        for line in f.readlines():
            line = line.strip()
            if line.startswith('module '):
                parts = line.split()
                module_name = parts[1]
                rstrip_nonword_pattern = r'\W+.*$'
                module_name = re.sub(rstrip_nonword_pattern, '', module_name)
                if bool(re.fullmatch(r'^\w+$', module_name)):
                    if module_name == module_guess:
                        return module_guess
                    if module_name:
                        best_guess = module_name
    if best_guess:
        return best_guess

    return ''



'''opencos.commands.waves - command handler for: eda waves ...

Note this command is handled differently than others (such as CommandSim),
it is generally run as simply

    > eda waves

and attempts to auto-find the newest waveform in your work directories. As
a result, no Tool is bound to CommandWaves (there is no CommandWavesVivado
handler).
'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

from pathlib import Path
import os

from opencos import eda_config
from opencos.eda_base import Command, Tool
from opencos.utils import markup_helpers
from opencos.util import Colors, debug, info, warning, import_class_from_string


class CommandWaves(Command):
    '''command handler for: eda waves'''

    command_name = 'waves'


    def __init__(self, config: dict):
        Command.__init__(self, config=config)
        self.args.update({
            'test-mode': False,
        })
        self.args_help.update({
            'test-mode': 'Do not run the command to open the located wave file, instead print' \
            + ' to stdout',
        })

        # process_tokens() will populate these:
        self.wave_dirs = [] # Targets presented as positional args
        self.wave_file = None # The one we'll be opening.
        self.unparsed_args = []


    def get_supported_wave_file_ext(self) -> list:
        '''Given the tools in config, return a list of all supported wave file extensions'''
        ret = []

        # tools to check, if this is a Tool class, then only check itself, otherwise
        # check everything from config.auto_tools_order:
        tools = []
        if isinstance(self, Tool):
            if _tool := getattr(self, '_TOOL', ''):
                tools.append(_tool) # one tool
        else:
            tools = self.config.get('auto_tools_order', {}).get('waves', [])

        for _tool in tools:
            if not _tool:
                continue

            _entry = self.config.get('tools', {}).get(_tool, {})
            if not _entry:
                continue

            if _tool not in self.config.get('tools_loaded', []):
                continue

            waves_file_ext_list = _entry.get('waves-file-ext', [])
            if isinstance(waves_file_ext_list, str):
                waves_file_ext_list = waves_file_ext_list.split()
            for ext in waves_file_ext_list:
                if ext not in ret:
                    ret.append(ext)

        return ret


    def get_wave_files_in_dirs(self, wave_dirs: list, quiet: bool = False) -> list:
        '''Returns list of all wave files give wave_dirs (list)'''

        supported_file_ext = self.get_supported_wave_file_ext()
        if not supported_file_ext:
            if not quiet:
                warning(f"No supported wave file extensions found, can't search: {wave_dirs}")
            return []

        if not wave_dirs:
            if not quiet:
                warning(f"No wave_dirs to search, for supported extensions: {supported_file_ext}")
            return []

        all_files = []
        debug(f"Looking for wave files in: {wave_dirs}")
        for d in wave_dirs:
            if not quiet:
                info(f"Looking for wave files below: {d}")
            for root, _, files in os.walk(d):
                for f in files:
                    if f and os.path.splitext(f)[1] in supported_file_ext:
                        if not quiet:
                            info(f"Found wave file: {Colors.byellow}{os.path.join(root, f)}")
                        all_files.append(os.path.join(root, f))
        return all_files


    def find_tool_used_on_wave_file(self) -> str:
        '''Try to find out which tool created self.wave_file

        For FST or VCD files, this doesn't matter, but for WLF files it matters
        '''
        wave_base_path = Path(self.wave_file).parent
        eda_output_config_fpath = wave_base_path / eda_config.EDA_OUTPUT_CONFIG_FNAME
        if eda_output_config_fpath.exists():
            data = markup_helpers.yaml_safe_load(str(eda_output_config_fpath))
            if tool := data.get('args', {}).get('tool', None):
                # This is set if an auto-tool was selected.
                return tool

        return ''


    def _get_child_handling_class(self) -> object: # pylint: disable=too-many-branches
        '''Returns a class handle of a child to process this, which should be a Tool class

        if no appropriate child is found, returns self.

        Uses self.wave_file, and assumes self.wave_file is already set.
        '''
        if isinstance(self, Tool):
            # We're already a tool handling class.
            return self

        assert self.wave_file, f'{self.wave_file=} not set, cant determine child class'

        ext = os.path.splitext(self.wave_file)[1]
        tool_guesses = []

        # Note - use the config.auto_tools_order so the CONFIG-YML tells us what order
        # to search for waveform viewers. Each viewer must have registered its file
        # extensions in config.tools.TOOLNAME.waves-files-ext (list)
        for _tool in self.config.get('auto_tools_order', {}).get('waves', []):
            if not _tool:
                continue

            _entry = self.config.get('tools', {}).get(_tool, {})
            if not _entry:
                continue

            if _tool not in tool_guesses and \
               _tool in self.config.get('tools_loaded', []) and \
               ext in _entry.get('waves-file-ext', []):
                tool_guesses.append(_tool) # append, need ordered list here.

        # There's an added wrinkle that if you have > 1 version of Questa available
        # in PATH (Questa, ModelsimAse, etc) we can't smartly pick a default waveform
        # viewer using the file type alone, if that was created under a different tool.
        # In this case, we should check to see if the self.wave_file directory has
        # an eda_output_config.yml file, and if so what tool was used to run the
        # simulation.
        prefer_waves_tool = None
        original_sim_tool = self.find_tool_used_on_wave_file()
        if original_sim_tool and \
           original_sim_tool in self.config.get('auto_tools_order', {}).get('waves', []):
            _entry = self.config.get('tools', {}).get(original_sim_tool, {})
            if ext in _entry.get('waves-file-ext', []):
                # We found the original, it supports "waves" commands, and it supports
                # this file extension.
                prefer_waves_tool = original_sim_tool


        debug(f'waves: {self.wave_file}: {tool_guesses=}')
        if not tool_guesses:
            warning(f'For wave file {self.wave_file} no tool found that can open it')
            return self

        if prefer_waves_tool and prefer_waves_tool in tool_guesses:
            tool_guesses = [prefer_waves_tool] # only try this one.

        for tool in tool_guesses:
            cls_str, cls, sco = None, None, None
            tool_cfg = self.config.get('tools', {}).get(tool, {})
            debug(f'waves: tool_guess {tool=}, {tool_cfg=}')
            if tool_cfg:
                cls_str = tool_cfg.get('handlers', {}).get(self.command_name, None)
            debug(f'waves: tool_guess {tool=}, {cls_str=}')
            if cls_str:
                cls = import_class_from_string(cls_str)
            debug(f'waves: tool_guess {tool=}, {cls=}')
            if cls and issubclass(cls, Command):
                info(f'For wave file {self.wave_file}, can use tool={tool}')
                debug(f'For wave file {self.wave_file}, tool={tool} has handler {cls}')
                sco = cls(config=self.config)
                return sco
            # warn if this 'tool' isn't setup properly.
            warning(f'No handler found for tool={tool} to process wave file: {self.wave_file}')

        debug(f'config -- tools_loaded: {self.config["tools_loaded"]}')
        return self


    def get_and_set_wave_file_from_args( # pylint: disable=too-many-branches
            self, quiet: bool = False
    ) -> str:
        '''Returns abspath of a wave file or empty str, will modify self.unparsed_args

        and sets self.wave_file and self.wave_dirs'''

        if self.wave_file:
            # If this was already set, then return it.
            return self.wave_file

        while self.unparsed_args:
            if os.path.isfile(self.unparsed_args[0]):
                if self.wave_file:
                    self.error(f"Already using positional arg FILE: {self.wave_file}; not",
                               f"sure what to do with another wave file: {self.unparsed_args[0]}")
                self.wave_file = os.path.abspath(self.unparsed_args[0])
                self.unparsed_args.pop(0)
                continue
            if os.path.isdir(self.unparsed_args[0]):
                try_dir = Path(self.unparsed_args[0]).resolve()
                if self.wave_file and not quiet:
                    info(f"Already using positional arg FILE: {self.wave_file};",
                              f"ignoring search directory: {try_dir}")
                elif try_dir not in self.wave_dirs:
                    self.wave_dirs.append(try_dir)
            self.warning_show_known_args()
            self.error(f"Didn't understand command arg/token: '{self.unparsed_args[0]}'",
                       "in CommandWaves")

        if not self.wave_file:
            if not quiet:
                info("Need to look for wave file")
            # we weren't given a wave file, so we will look for one!
            if not self.wave_dirs and os.path.isdir(self.args['eda-dir']):
                try_dir = Path(self.args['eda-dir']).resolve()
                if try_dir not in self.wave_dirs:
                    self.wave_dirs.append(str(try_dir))
            if not self.wave_dirs:
                self.wave_dirs.append(os.getcwd())
            all_files = self.get_wave_files_in_dirs(self.wave_dirs)
            if len(all_files) > 1:
                all_files.sort(key=os.path.getmtime)
                if not quiet:
                    info(f"Choosing: {all_files[-1]} (newest)")
            if all_files:
                self.wave_file = all_files[-1]
            else:
                self.error(f"Couldn't find any wave files below: {','.join(self.wave_dirs)}")

        self.wave_file = os.path.abspath(self.wave_file)
        return self.wave_file


    def process_tokens( # pylint: disable=too-many-branches,too-many-statements
            self, tokens: list, process_all: bool = True,
            pwd: str = os.getcwd()
    ) -> list:

        self.unparsed_args = Command.process_tokens(
            self, tokens=tokens, process_all=False, pwd=pwd
        )

        if self.args['test-mode']:
            self.exec = self._test_mode_exec

        self.get_and_set_wave_file_from_args()

        # Since we're already in a class, and not back in eda.py, we can
        # determine which child handling class should service self.wave_file based
        # on the auto tool order and the extension.
        sco = self._get_child_handling_class()
        if sco is None or not isinstance(sco, Tool):
            self.error(f'Could not find a suitable tool to process file: {self.wave_file}')
            return []

        # Create the child handling class, parse the same original tokens, set
        # somethings ahead of time.
        sco.wave_dirs = self.wave_dirs
        sco.wave_file = self.wave_file
        sco.unparsed_args = Command.process_tokens(
            sco, tokens=tokens, process_all=False, pwd=pwd
        )
        if self.args['test-mode']:
            sco.exec = self._test_mode_exec

        debug(f'We will use {sco=} {type(sco)=} for {self.wave_file=}')
        sco.do_it()
        return sco.unparsed_args


    def _test_mode_exec( # pylint: disable=unused-argument
            self, work_dir: str,
            command_list: list,
            **kwargs
    ) -> None:
        '''Override for Command.exec if arg --test-mode was set, does not run

        the command_list, instead prints to stdout'''

        info(f'waves.py: {Colors.off}{Colors.bcyan}test_mode exec{Colors.info} stdout:',
             f'{Colors.off}{" ".join(command_list)}{Colors.info};   ({work_dir=}')



#!/usr/bin/env python3

'''opencos.eda_deps_sanitize is an executable script

Usage:
    eda_deps_sanitize --dir=<path>

Will print santized JSON data for the DEPS file found in --dir=<path>
'''

import argparse
import os
import sys

from pathlib import Path

from opencos import util
from opencos.deps import deps_file
from opencos.utils import status_constants


def run(*args) -> (int, str):
    '''Runs the DEPS sanitizer, prints results to stdout'''


    bool_kwargs = util.get_argparse_bool_action_kwargs()

    parser = argparse.ArgumentParser(
        prog='opencos eda_deps_sanitize', add_help=True, allow_abbrev=False
    )

    parser.add_argument('--yaml', **bool_kwargs,
                        help='Print output as YAML text, otherwise default is JSON text')
    parser.add_argument('dir', type=str, default=str(Path('.')),
                        help='Directory to look for DEPS.[markup] file')

    try:
        parsed, unparsed = parser.parse_known_args(list(args) + [''])
        unparsed = list(filter(None, unparsed))
    except argparse.ArgumentError:
        return 1, f'problem attempting to parse_known_args for {args=}'

    deps_path = parsed.dir
    if os.path.isfile(deps_path):
        deps_path, _ = os.path.split(deps_path)

    try:
        my_depsfile_obj = deps_file.DepsFile(None, deps_path, {})
        if not my_depsfile_obj.deps_file:
            return status_constants.EDA_DEPS_FILE_NOT_FOUND, f'No DEPS markup file at {parsed.dir}'

        ret_str = my_depsfile_obj.str_sanitized_markup(as_yaml=parsed.yaml)
        rc = util.get_return_code()
        return rc, ret_str
    except Exception as e:
        rc = 1, str(e)

    return 0, ''


def main() -> None:
    '''calls sys.exit(), main entrypoint'''
    args = []
    if len(sys.argv) > 1 and not args:
        args = sys.argv[1:]

    rc, deps_str = run(*args)
    print(deps_str)
    sys.exit(rc)


if __name__ == '__main__':
    main()



'''opencos.utils.ssh_helper -- remote EDA execution via SSH

Invoked when `eda` is called with --ssh <user>@<host>[:/optional/path].

The :/path is optional.  When omitted (or when you want a neutral scratch space),
the remote sandbox is placed under the user's home directory on the remote host:
  ~/eda.work/<uuid>.ssh/
When :/path is given it is used as the base instead:
  /path/eda.work/<uuid>.ssh/
Either way the sandbox is ephemeral -- created, used, and deleted per job.

High-level flow:
  1. parse_ssh_arg()           -- split "user@host[:/path]" into (user, host, path|None)
  2. _get_remote_home()        -- resolve $HOME on remote (used when no path given)
  3. remote_check_eda()        -- verify `eda` is reachable on the remote host
  4. remote_check_tool()       -- verify the requested --tool exe is on the remote
  5. _stage_files()            -- collect source files → flat local staging dir via ExportHelper
  6. _ship_to_remote()         -- tar-pipe staging dir to a fresh ephemeral remote job dir
  7. run_eda_remote()          -- cd remote_job_dir && eda <args>; stream output
  8. _fetch_results()          -- tar-pipe remote eda.work/ back to local eda.work/
  9. _cleanup_remote()         -- ssh rm -rf remote_job_dir  (ephemeral, always removed)
'''

import json
import os
import re
import shlex
import subprocess
import tempfile
import uuid

from opencos import util, files, eda_config, __version__
from opencos.util import Colors
from opencos.export_helper import ExportHelper
from opencos.utils import subprocess_helpers
from opencos.utils.status_constants import EDA_SSH_ERROR_CODE, EDA_COMMAND_OR_ARGS_ERROR, \
    EDA_COMMAND_MISSING_TOP


# ---------------------------------------------------------------------------
# Parsing
# ---------------------------------------------------------------------------

SSH_ARG_RE = re.compile(
    r'^(?P<user>[^@]+)@(?P<host>[^:/]+)(?::(?P<path>.+))?$'
)

SSH_EXE = files.safe_shutil_which('ssh')
TAR_EXE = files.safe_shutil_which('tar')

MIN_VERSION = '0.4.4'

# hold path to remote eda, remote eda version, remote known tools
# per user@host entry:
REMOTE_CACHE = {}

def _init_remote_cache(user:str, host: str) -> str:
    '''Creates a REMOTE_CACHE entry if it doesn't already exist'''
    key = f'{user}@{host}'
    if key not in REMOTE_CACHE:
        REMOTE_CACHE[key] = {
            'eda': {
                'exe': '',
                'version': '',
            },
            'tools': {
            },
        }
    return key

def _version_ok(ver: str) -> bool:
    '''Checks if version can handle ssh remote flow'''
    if not ver:
        return False

    parts = ver.split('.')
    if parts and len(parts) >= 3:
        if any((int(parts[0]) > 0,
                int(parts[1]) > 4,
                int(parts[1]) == 4 and int(parts[2]) >= 4)):
            # 1+.x.x, 0.5.x, or 0.4.4+
            return True

    return False


def parse_ssh_arg(ssh_arg: str) -> tuple:
    '''Parse "user@host[:/some/path]" into (user, host, path_or_None).

    The :/path portion is optional.  When omitted, path is returned as None
    and the caller should use the remote home directory instead.

    Raises ValueError with a human-readable message on bad format.
    '''
    m = SSH_ARG_RE.match(ssh_arg.strip())
    if not m:
        raise ValueError(
            f'--ssh value must be in the form user@host or user@host:/path, got: {ssh_arg!r}'
        )
    user = m.group('user')
    host = m.group('host')
    path = m.group('path')  # None when no :/path was given
    if path is not None and not path.startswith('/'):
        raise ValueError(
            f'--ssh path must be absolute (start with /), got: {path!r}'
        )
    return user, host, path


# ---------------------------------------------------------------------------
# SSH helpers
# ---------------------------------------------------------------------------

def _prepend_setup_cmd(cmd: str, *setup_cmds) -> str:
    '''Prepends the cmd with setup_cmds'''
    if not setup_cmds:
        return cmd
    return ' && '.join(list(filter(None, setup_cmds)) + [cmd])

def _ssh_base(user: str, host: str) -> list:
    '''Returns the base ssh command list for a given user@host.'''
    if not SSH_EXE:
        util.error('"ssh" not present in PATH, cannot use "eda --ssh=SSH_VALUE")',
                   error_code=EDA_SSH_ERROR_CODE)
        return []
    return [
        SSH_EXE,
        '-o', 'BatchMode=yes',
        '-o', 'ConnectTimeout=10',
        f'{user}@{host}',
    ]


def _ssh_cmd(user: str, host: str, remote_cmd: str) -> list:
    '''Build an SSH command that runs remote_cmd via "bash -lc" so ~/.bashrc is sourced.

    This ensures tools added to PATH in .bashrc are available on the remote.
    '''
    assert _ssh_base
    return _ssh_base(user, host) + [f'bash -lc {_shell_quote(remote_cmd)}']


def _run_local(
        cmd: list, *, capture: bool = False, check: bool = True
) -> subprocess.CompletedProcess:
    '''Run a local command. When capture=False output streams to the terminal.'''
    util.debug(f'ssh_helper: running local cmd: {" ".join(str(c) for c in cmd)}')
    kwargs = subprocess_helpers.get_kwargs(work_dir='', shell=False, is_subprocess_popen=False)
    if capture:
        kwargs['capture_output'] = True
        kwargs['text'] = True
    # Note - we do not have subprocess_helpers PID management for subprocess.run where
    # we capture text from STDOUT.
    return subprocess.run(cmd, check=check, **kwargs)


def remote_ensure_path(user: str, host: str, path: str) -> None:
    '''Create <path> on the remote if it does not exist.

    Raises RuntimeError if the remote is unreachable or mkdir fails.
    '''
    util.info(f'ssh_helper: ensuring remote path exists: {user}@{host}:{path}')
    cmd = _ssh_cmd(user, host, f'mkdir -p {path!r}')
    try:
        _run_local(cmd, capture=True)
    except subprocess.CalledProcessError as e:
        util.error(
            f'ssh_helper: cannot reach {user}@{host} or create {path}: {e}',
            error_code=EDA_SSH_ERROR_CODE
        )


def _get_remote_home(user: str, host: str) -> str:
    '''Return the absolute home directory path on the remote host.

    Runs "echo $HOME" via bash -lc so the shell expands the variable.
    Raises RuntimeError if the remote is unreachable or $HOME is empty.
    '''
    cmd = _ssh_cmd(user, host, 'echo $HOME')
    try:
        result = _run_local(cmd, capture=True)
        home = result.stdout.strip()
    except subprocess.CalledProcessError as e:
        util.error(
            f'ssh_helper: cannot reach {user}@{host} to determine $HOME: {e}',
            error_code=EDA_SSH_ERROR_CODE
        )
    if not home or not home.startswith('/'):
        util.error(
            f'ssh_helper: unexpected $HOME value on {user}@{host}: {home!r}',
            error_code=EDA_SSH_ERROR_CODE
        )
    return home


def remote_check_eda(user: str, host: str, setup_cmd: str = '') -> str:
    '''Returns the remote path to the eda executable, or empty string if not found.'''
    ret = ''
    which_cmd = _prepend_setup_cmd('which eda 2>/dev/null || echo ""', setup_cmd)
    cmd = _ssh_cmd(user, host, which_cmd)
    try:
        result = _run_local(cmd, capture=True)
        ret = _last_nonempty_line(result.stdout)
    except subprocess.CalledProcessError:
        ret = ''

    key = _init_remote_cache(user, host)
    REMOTE_CACHE[key]['eda']['exe'] = ret
    return ret


def remote_eda_version(user: str, host: str, setup_cmd: str = '') -> str:
    '''Get remote eda version'''
    eda_ver_cmd = _prepend_setup_cmd('eda --version', setup_cmd)
    cmd = _ssh_cmd(user, host, eda_ver_cmd)
    ver = ''
    try:
        result = _run_local(cmd, capture=True)
        ver = _last_nonempty_line(result.stdout)
        if not ver:
            util.warning(
                f'ssh_helper: {user}@{host} -- eda version: {ver=} -- from {result.stdout}'
            )
    except subprocess.CalledProcessError as e:
        util.warning(f'ssh_helper: {user}@{host} -- eda version: {ver=} -- has error ({e})')
        ver = ''

    if not ver:
        util.warning(f'ssh_helper: {user}@{host} -- eda version: {ver=} -- from {result.stdout}')
        return ver

    parts = ver.split()
    if len(parts) < 2:
        return ver

    ver = parts[1]
    key = _init_remote_cache(user, host)
    REMOTE_CACHE[key]['eda']['version'] = ver
    util.debug(f'ssh_helper: {user}@{host} -- eda version: {ver}')
    return ver


def get_remote_tools(user: str, host: str, setup_cmd: str = '') -> dict:
    '''
    Runs "eda show-tools" on remote (if opencos-eda >= 0.4.4) to get JSON

    data of all tools/commands supported. If show-tools fails, then we will use
    "eda --help" and attempt to parse from STDOUT.

    Also sets in REMOTE_KNOWN_TOOLS
    '''
    key = _init_remote_cache(user, host)

    data = None
    if 'tools' in REMOTE_CACHE[key].get('tools', {}):
        data = REMOTE_CACHE[key]['tools']
    else:
        ver = remote_eda_version(user, host, setup_cmd)
        show_tools_works = False
        if ver:
            show_tools_works = _version_ok(ver)
            if __version__ != ver:
                util.warning(
                    f'Local "eda" version is {Colors.byellow}{__version__}{Colors.yellow}',
                    f', remote is {Colors.byellow}{ver}')

        if show_tools_works:
            try:
                cmd = _ssh_cmd(
                    user, host,
                    _prepend_setup_cmd('eda show-tools --quiet --json-to-stdout', setup_cmd)
                )
                result = _run_local(cmd, capture=True)
                show_tools_data = json.loads(result.stdout)
            except Exception as e:
                util.error(f'Cannot get tools information on remote {key}: {cmd} ({e})')

            if 'tools' in show_tools_data:
                REMOTE_CACHE[key]['tools'] = show_tools_data['tools'].copy()
                data = REMOTE_CACHE[key]['tools']
            else:
                util.error(
                    f'Cannot find tool information from: {key}: {cmd}; got -- {show_tools_data}'
                )

        else:
            util.error(f'SSH requires eda >= 0.4.4 to get tool information on remote {key}')
            data = None

    return data


def remote_check_tool(user: str, host: str, tool_name: str, setup_cmd: str = '') -> str:
    '''Returns the remote exe path for tool_name (first exe name in PATH), or empty string.

    Uses a simple heuristic: the tool exe has the same name as the tool key.
    For tools like slang_yosys, we check "yosys".
    '''
    # TODO(drew): we can likely use `eda show-tools [--args]` to get this in a JSON
    # usable format from the remote host

    data = get_remote_tools(user, host, setup_cmd)
    exe = None
    if tool_name in data:
        exe = data[tool_name].get('path', '')

    if not exe:
        util.error(f'tool {tool_name} does not appear to exist on remote {user}@{host}')
        return ''

    return exe


def _last_nonempty_line(text: str) -> str:
    '''Return the last non-empty line from text (setup commands may produce leading output).'''
    for line in reversed(text.splitlines()):
        stripped = line.strip()
        if stripped:
            return stripped
    return ''


# ---------------------------------------------------------------------------
# File staging (ExportHelper → flat local dir)
# ---------------------------------------------------------------------------

def _get_target_name(cmd_obj) -> str:
    '''Gets member 'target' or args['top'], returns empty str otherwise'''
    target = getattr(cmd_obj, 'target', '')
    if not target:
        target = cmd_obj.args.get('top', '')
    return target

def _stage_files(cmd_obj, staging_dir: str) -> int:
    '''Collect all source files into staging_dir using ExportHelper.

    ExportHelper creates a flat self-contained directory: all source files by
    basename, with `include` paths rewritten to be relative (no subdirectory
    prefixes), plus a new DEPS.yml with incdirs:[.] and the target key.

    Returns 0 on success, non-zero on error.
    '''
    target = _get_target_name(cmd_obj)
    if not target:
        target = cmd_obj.args.get('top', '')
        util.warning(f'ssh_helper: target not found, using "top" value: {target}')

    if not target:
        return util.error(
            'ssh_helper: cannot determine target for staging (self.target is empty).',
            error_code=EDA_COMMAND_OR_ARGS_ERROR
        )

    if not cmd_obj.args.get('top', ''):
        return util.error(
            'ssh_helper: --top not set after DEPS parsing, cannot stage files.'
            ' Ensure your DEPS.yml target has a `top:` field.',
            error_code=EDA_COMMAND_MISSING_TOP
        )

    util.info(f'ssh_helper: staging files for {target=} → {staging_dir}')

    # TODO(drew): confirm this works, ExportHelper may need allow-listed args per tool
    export = ExportHelper(
        cmd_design_obj=cmd_obj,
        eda_command=cmd_obj.command_name,
        out_dir=staging_dir,
        target=target,
    )
    # staging_dir already exists (created by TemporaryDirectory); ExportHelper's
    # create_out_dir() would try to mkdir it, but write_files_to_out_dir() and
    # create_deps_yml_in_out_dir() work fine with an existing directory.
    export.write_files_to_out_dir()
    export.create_deps_yml_in_out_dir()

    staged = os.listdir(staging_dir)
    util.info(f'ssh_helper: staged {len(staged)} files: {staged}')
    return 0


# ---------------------------------------------------------------------------
# File transfer (tar+pipe over SSH)
# ---------------------------------------------------------------------------

def _ship_to_remote(user: str, host: str, staging_dir: str, remote_job_dir: str) -> int:
    '''Ship staging_dir contents to a fresh remote_job_dir via tar+ssh pipe.

    Creates remote_job_dir on the remote host, then streams a tar archive of
    staging_dir over SSH so no intermediate copy is needed.

    Returns 0 on success, non-zero on error.
    '''
    util.info(f'ssh_helper: shipping files to {user}@{host}:{remote_job_dir}')
    try:
        remote_ensure_path(user, host, remote_job_dir)
    except RuntimeError as e:
        return util.error(str(e), error_code=EDA_SSH_ERROR_CODE)

    if not TAR_EXE:
        util.error(
            '"tar" is not present in PATH, cannot use ssh to send file to remote',
            error_code=EDA_SSH_ERROR_CODE
        )

    # --no-xattrs: macOS bsdtar would otherwise embed LIBARCHIVE.xattr.* PAXHEADER entries
    # for files that carry extended attributes (e.g. com.apple.provenance), which causes
    # harmless but noisy "Ignoring unknown extended header keyword" warnings when Linux tar
    # extracts them.  The flag is a no-op on Linux tar, so safe to pass always.
    tar_cmd = [TAR_EXE, '--no-xattrs', '-C', staging_dir, '-czf', '-', '.']
    # TODO(drew): might need to get the remote's TAR_EXE, currently assumes 'tar'
    ssh_cmd = _ssh_base(user, host) + [f'tar -C {remote_job_dir!r} -xzf -']

    # TODO(drew): lost support for --subprocess-timeout and --subprocess-progress-timeout.
    # pylint: disable=consider-using-with
    tar_proc_kwargs = subprocess_helpers.get_kwargs(work_dir='')
    tar_proc_kwargs['stdout'] = subprocess.PIPE
    tar_proc = subprocess.Popen(tar_cmd, **tar_proc_kwargs)
    subprocess_helpers.add_running_parent_pid(tar_proc.pid)

    ssh_proc_kwargs = subprocess_helpers.get_kwargs(work_dir='', is_subprocess_popen=False)
    ssh_proc_kwargs.update({
        'stdin': tar_proc.stdout,
    })
    ssh_proc = subprocess.run(ssh_cmd, **ssh_proc_kwargs, check=False)
    tar_proc.stdout.close()
    tar_proc.wait()
    subprocess_helpers.remove_completed_parent_pid(tar_proc.pid)

    if tar_proc.returncode != 0 or ssh_proc.returncode != 0:
        return util.error(
            f'ssh_helper: file transfer failed '
            f'(tar rc={tar_proc.returncode}, ssh rc={ssh_proc.returncode})',
            error_code=EDA_SSH_ERROR_CODE
        )
    util.info('ssh_helper: files shipped successfully.')
    return 0


def _fetch_results(user: str, host: str, remote_job_dir: str, local_eda_work: str) -> None:
    '''Fetch job artifacts from the remote back to local_eda_work.

    Uses tar+pipe to bring back the entire remote eda.work/ tree.  Because the remote
    job directory is an ephemeral sandbox (created fresh for this job, removed afterwards),
    eda.work/ contains only the results from this specific run — there are no stale files.

    After the transfer, reads the local artifacts.json (if present) to report what was
    produced.
    '''
    remote_work = f'{remote_job_dir}/eda.work'

    # Check whether eda.work was created at all.
    check = _run_local(
        _ssh_base(user, host) + [
            f'bash -lc {_shell_quote("[ -d " + remote_work + " ] && echo exists || echo absent")}'
        ],
        capture=True, check=False
    )
    if 'absent' in check.stdout:
        util.warning('ssh_helper: no remote eda.work/ found — nothing to fetch.')
        return

    util.info(f'ssh_helper: fetching results from {user}@{host}:{remote_work}')
    os.makedirs(local_eda_work, exist_ok=True)

    # Stream the entire eda.work/ tree back via tar+pipe.
    ssh_send_cmd = _ssh_base(user, host) + [f'tar -C {remote_work!r} -czf - .']
    # TODO(drew): lost support for --subprocess-timeout and --subprocess-progress-timeout.
    # pylint: disable=consider-using-with
    ssh_proc_kwargs = subprocess_helpers.get_kwargs(work_dir='')
    ssh_proc_kwargs['stdout'] = subprocess.PIPE
    ssh_proc = subprocess.Popen(ssh_send_cmd, **ssh_proc_kwargs)
    subprocess_helpers.add_running_parent_pid(ssh_proc.pid)

    tar_proc_kwargs = subprocess_helpers.get_kwargs(work_dir='', is_subprocess_popen=False)
    tar_proc_kwargs.update({
        'stdin': ssh_proc.stdout,
    })
    tar_proc = subprocess.run(
        ['tar', '-C', local_eda_work, '-xzf', '-'],
        check=False, **tar_proc_kwargs
    )

    ssh_proc.stdout.close()
    ssh_proc.wait()
    subprocess_helpers.remove_completed_parent_pid(ssh_proc.pid)
    if tar_proc.returncode != 0:
        util.warning(f'ssh_helper: eda.work/ fetch had errors (rc={tar_proc.returncode}).')


def _cleanup_remote(user: str, host: str, remote_job_dir: str) -> None:
    '''Remove the ephemeral remote job directory.'''
    if not remote_job_dir:
        return
    util.info(f'ssh_helper: cleaning up remote sandbox: {user}@{host}:{remote_job_dir}')
    _run_local(
        _ssh_cmd(user, host, f'rm -rf {remote_job_dir!r}'),
        capture=True, check=False
    )


# ---------------------------------------------------------------------------
# Remote eda invocation
# ---------------------------------------------------------------------------

def _build_remote_eda_args(
        eda_non_positional_args: list, target_key: str,
        run_no_color: bool = False
) -> list:
    '''Strip --ssh from original_args; normalise target token to target_key.

    Handles "--ssh=foo" and "--ssh foo" styles.

    The remote staging dir has a flat DEPS.yml — files and the target are referenced
    by basename only.  Replace any path-qualified target token with just target_key:
      "path/to/DEPS.yml::a01"    (DEPS-file-qualified)
      "A01_some_dir/a01"         (directory-prefixed, e.g. from eda multi)
    '''

    out = []
    skip_next = False
    for tok in eda_non_positional_args + [target_key]:
        if skip_next:
            skip_next = False
            continue
        if run_no_color and tok in ('--color', '--no-color'):
            continue # strip
        if tok in ('--ssh', '--ssh-setup'):
            skip_next = True
            continue
        if tok.startswith('--ssh=') or tok.startswith('--ssh-setup='):
            continue
        # Replace path-qualified target tokens with just the target key.
        # Leaving this, but we've already stripped the positional args
        # hence the naming eda_non_positional_args:
        if target_key and not tok.startswith('-') and (
            tok == target_key or
            tok.endswith('/' + target_key) or
            tok.endswith('::' + target_key)
        ):
            out.append(target_key)
            continue
        out.append(tok)

    if run_no_color:
        out.append('--no-color')
    return out


def run_eda_remote(
        user: str, host: str, remote_path: str,
        remote_eda_args: list,
        remote_eda_exe: str = 'eda',
        setup_cmd: str = '',
        timeout: int = 0,
        progress_timeout: int = 0
) -> int:
    '''Run eda on the remote host. Streams stdout/stderr to local terminal.

    Returns the remote eda exit code.
    '''
    eda_cmd = ' '.join([remote_eda_exe] + [_shell_quote(a) for a in remote_eda_args])
    parts = []
    if setup_cmd:
        parts.append(setup_cmd)
    parts.append(f'cd {remote_path!r}')
    parts.append(eda_cmd)
    remote_shell_cmd = ' && '.join(parts)

    ssh_cmd = _ssh_cmd(user, host, remote_shell_cmd)

    util.info(f'ssh_helper: running on {user}@{host}: {eda_cmd}')
    util.info(f'ssh_helper: remote working directory: {remote_path}')

    _, _, rc = subprocess_helpers.subprocess_run_background(
        work_dir='',
        command_list = ssh_cmd,
        background=False,
        shell=False, tee_fpath='',
        timeout=timeout, timeout_str='',
        progress_timeout=progress_timeout, progress_timeout_str='',
        force_no_color=False # attempt to keep colors for background=False
    )
    return rc


def _shell_quote(s: str, shlex_quote: bool = False) -> str:
    '''Minimal shell quoting: wraps in single quotes if the token contains spaces or specials.'''
    if shlex_quote:
        return shlex.quote(s)
    if re.search(r'[\s\'\"\\$`&|<>(){}]', s):
        return "'" + s.replace("'", "'\\''") + "'"
    return s


def _get_modified_args_list( # pylint: disable=too-many-branches
        args: dict, modified_args: dict, warn: bool = True, filter_args=None
) -> list:
    '''Givens args and modified_args (dict), return a list of str suitable for

    command line args. Reminder that modified args is {key_name: bool} (value True if modified)
    '''
    ret = []
    for arg, modified in modified_args.items():
        if not arg or not modified:
            continue
        if arg in filter_args:
            continue


        value = args[arg]
        if value is None:
            util.error(f'arg {arg} has been modified to None, cannot be re-applied as',
                       'command line arg')
        elif isinstance(value, dict):
            if warn:
                util.warning(f'arg {arg} is dict, {value=} cannot be re-applied as command line')
        elif isinstance(value, list):
            for x in value:
                ret.append(f'--{arg}={x}')
        elif isinstance(value, bool):
            if value:
                ret.append(f'--{arg}')
            else:
                ret.append(f'--no-{arg}')
        elif isinstance(value, int):
            ret.append(f'--{arg}={value}')
        elif isinstance(value, str):
            ret.append(f'--{arg}={_shell_quote(value)}')
        else:
            util.error(f'arg {arg} has {value=} {type(value)=}, cannot re-apply this type for',
                       'command line arg')


    return ret


def _get_eda_modified_args_list( # pylint: disable=dangerous-default-value
        cmd_obj, warn: bool = True, filter_args=None
) -> list:
    '''Examines cmd_obj.modified_args, eda_config, and util, and returns a list of str args

    suitable to re-apply args via ssh on remote host, this way expanded items
    like --input-file are not passed to remote.

    Note this does not include the eda COMMAND [SUBCOMMANDS]
    '''

    # NOTE(drew): even this is not perfect, would need to get various util.args
    # that were modified, loaded --env-file(s)
    ret = []

    if util.env_files_loaded:
        util.warning(
            '--env-file(s) have been loaded, these are not supported by SSH',
            f'and will not be sent to the remote host as eda args: {util.env_files_loaded}'
        )

    # Re-create the eda_config, util, and cmd_obj args.
    ret.extend(eda_config.get_config_yml_args_for_flist())
    ret.extend(
        _get_modified_args_list(
            util.args, util.modified_args, warn=warn, filter_args=filter_args
        )
    )
    ret.extend(
        _get_modified_args_list(
            cmd_obj.args, cmd_obj.modified_args, warn=warn,
            filter_args=list(util.modified_args.keys())
        )
    )

    return ret



# ---------------------------------------------------------------------------
# Main entry point (called from eda.py after process_tokens())
# ---------------------------------------------------------------------------

def run_ssh_command( # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        cmd_obj, local_cwd: str
) -> int:
    '''Ship source files, run eda remotely, fetch artifacts, clean up remote sandbox.

    Called from eda.py after process_tokens() has populated the file lists (files_sv,
    files_v, incdirs, defines, etc.) but before local tool execution.

    cmd_obj  -- CommandDesign instance with populated file lists
    local_cwd -- local working directory when eda was invoked (used for result placement)

    Returns a bash-style exit code (0 = pass).
    '''
    config = cmd_obj.config
    ssh_arg = config.get('ssh', '')
    ssh_setup = config.get('ssh_setup', '')

    # Note - the eda_orignal_args are literally what was on the command line,
    # so this does not include --input-file items resolved/expanded. So if you
    # created ./my_eda_ssh.f, and ran: eda sim mytarget -f=my_eda_ssh.f, it
    # would not work. We instead re-create from util, eda_config, cmd_obj modified_args.
    original_args = config.get('eda_original_args', [])

    # Get the modified args from
    cmd_obj_modified_args = _get_eda_modified_args_list(cmd_obj)
    tool = cmd_obj.args.get('tool', '') or config.get('ssh_tool', '')

    if not tool:
        return util.error(
            '--ssh requires --tool to be specified explicitly '
            '(remote tool discovery is not performed locally). '
            'Example: eda sim --tool vivado --ssh user@host',
            error_code=EDA_SSH_ERROR_CODE
        )

    try:
        user, host, explicit_path = parse_ssh_arg(ssh_arg)
    except ValueError as e:
        return util.error(str(e), error_code=EDA_SSH_ERROR_CODE)

    if explicit_path:
        remote_base_path = explicit_path
    else:
        try:
            remote_home = _get_remote_home(user, host)
        except RuntimeError as e:
            return util.error(str(e), error_code=EDA_SSH_ERROR_CODE)
        remote_base_path = remote_home

    util.info(
        f'ssh_helper: {Colors.bcyan}--ssh{Colors.off_green} mode:'
        f' {Colors.byellow}{user}@{host}'
        + (f'{Colors.off_green} (base: {remote_base_path})' if explicit_path
           else f'{Colors.off_green} (base: {remote_base_path} from $HOME)')
    )

    # --- Verify remote eda ---
    remote_eda_exe = remote_check_eda(user, host, setup_cmd=ssh_setup)
    if not remote_eda_exe:
        return util.error(
            f'ssh_helper: `eda` was not found in PATH on {user}@{host}.'
            ' Please install opencos-eda on the remote host'
            ' (e.g. uv tool install opencos-eda, pip install opencos-eda, or activate'
            ' the appropriate venv).',
            error_code=EDA_SSH_ERROR_CODE
        )
    util.info(f'ssh_helper: remote eda found: {remote_eda_exe}')

    # --- Verify remote tool ---
    tool_name = tool.split('=')[0]
    remote_tool_exe = remote_check_tool(user, host, tool_name, setup_cmd=ssh_setup)
    if remote_tool_exe:
        util.info(
            f'ssh_helper: remote {Colors.off_bgreen}{tool_name}{Colors.info}'
            f'{Colors.off_green} found: {remote_tool_exe}'
        )
    else:
        util.warning(
            f'ssh_helper: tool {tool_name!r} was not found on {user}@{host}.'
            ' The remote eda run may fail if the tool is not in PATH.'
        )

    # --- Create unique ephemeral remote job directory ---
    job_id = uuid.uuid4().hex[:12]
    remote_job_dir = f'{remote_base_path}/eda.work/{job_id}.ssh'
    util.info(f'ssh_helper: remote job dir: {remote_job_dir}')

    # --- Stage files locally, ship to remote, run, fetch, clean up ---
    with tempfile.TemporaryDirectory(prefix='eda_ssh_stage_') as staging_dir:
        rc = _stage_files(cmd_obj, staging_dir)
        if rc != 0:
            return rc

        rc = _ship_to_remote(user, host, staging_dir, remote_job_dir)
        if rc != 0:
            _cleanup_remote(user, host, remote_job_dir)
            return rc

    target_key = _get_target_name(cmd_obj)

    commands = []
    if c := getattr(cmd_obj, 'command_name', ''):
        commands.append(c)
    if c := getattr(cmd_obj, 'subcommand', ''):
        commands.append(c)

    if not commands:
        util.error(f'Could not find the commands to run: {original_args=}',
                   f'fished out {cmd_obj_modified_args=} and {target_key=}, but {commands=}')
    if not target_key:
        util.error(f'Could not find the exported target to run: {original_args=}',
                   f'fished out {cmd_obj_modified_args=} and {commands=}, but {target_key=}')

    remote_args = commands + _build_remote_eda_args(
        cmd_obj_modified_args, target_key, run_no_color=False
    )

    util.debug(f'ssh_helper.py: run_eda_remote -- {remote_args=}')

    rc = run_eda_remote(
        user=user,
        host=host,
        remote_path=remote_job_dir,
        remote_eda_args=remote_args,
        remote_eda_exe=remote_eda_exe,
        setup_cmd=ssh_setup,
        timeout=cmd_obj.args.get('subprocess-timeout', 0),
        progress_timeout=cmd_obj.args.get('subprocess-progress_timeout', 0)
    )

    local_eda_work = os.path.join(local_cwd, 'eda.work')
    _fetch_results(user, host, remote_job_dir, local_eda_work)

    _cleanup_remote(user, host, remote_job_dir)

    if rc == 0:
        util.info(
            f'ssh_helper: {Colors.off_bgreen}remote eda completed successfully.'
            f'{Colors.info} Results in local eda.work/'
        )
    else:
        util.error(
            f'ssh_helper: remote eda exited with code {rc}.'
            ' Check output above for details.',
            error_code=EDA_SSH_ERROR_CODE
        )

    return rc



'''opencos.commands.deps_help - command handler for: eda deps_help [args]

Note this command is handled differently than others (such as CommandSim),
it is generally run as simply

    > eda deps_help
    > eda deps_help --verbose
    > eda deps_help --help

uses no tools and will print a help text regarding DEPS markup files to stdout.
'''

# pylint: disable=line-too-long

import os
import re

from opencos.eda_base import Command
from opencos import util
from opencos.util import Colors

def get_deps_md_file() -> str:
    '''Tries to get docs/DEPS.md from our pypackage dist'''
    opencos_dir, _ = os.path.split(util.__file__)

    # Try to get it from site-packages dir, which should have docs/ alongside
    # commands/ and tools/
    filename = os.path.join(opencos_dir, 'docs', 'DEPS.md')
    if os.path.isfile(filename):
        return filename

    # If you're running directly from the git checkout dir, you won't be getting
    # this dist, so it's not in opencos/docs, it will simply be in ./docs
    filename = os.path.join(opencos_dir, '..', 'docs', 'DEPS.md')
    if os.path.isfile(filename):
        return filename
    return ''

def get_deps_md_contents() -> str:
    '''Tries to get the docs/DEPS.md file and returns the str contents

    This also performs some limited colorization of markdown and YAML
    (assuming util was not disabled with --no-color)
    '''
    filename = get_deps_md_file()
    if not filename:
        return ''

    def make_byellow(match):
        '''Used by re.sub to wrap the match with bold yellow and return to normal yellow'''
        return f'{Colors.byellow}{match.group(0)}{Colors.off_yellow}'

    lines = []
    with open(filename, encoding='utf-8') as f:
        for line in f.readlines():

            if line.startswith('# '):
                # colors for markdown headings
                line = f'{Colors.bgreen}{line}{Colors.off_yellow}'
            elif line.startswith('## '):
                # colors for markdown headings
                line = f'{Colors.bcyan}{line}{Colors.off_yellow}'

            elif '#' in line:
                # colors for comments
                line = line.replace('#', f'{Colors.off_cyan}#') + Colors.yellow

            # colors for starting a line with:
            #  key: value
            #  - key : value
            # try to make "key:"  or "- key:" as bold:
            line = re.sub(
                r'^( *\-? ?[^ ]+):', make_byellow, line
            )

            lines.append(line)

    return ''.join(lines)


BASIC_DEPS_HELP = f'''
{Colors.yellow}
Note: you can run with one of: {Colors.cyan}--verbose, --help, --debug{Colors.yellow} to show full
schema supported, or {Colors.cyan}--no-color{Colors.yellow} to avoid printing this text with colors.

{Colors.green}--------------------------------------------------------------------{Colors.yellow}

  What is a {Colors.byellow}DEPS.yml{Colors.off}{Colors.yellow} file and why does `eda` use this?
  - {Colors.byellow}DEPS.yml{Colors.off}{Colors.yellow} is a fancy filelist.
  - Used to organize a project into "targets", a tool can run on a "target".
  - Allows for more than just source files attached to a "target".
    -- incdirs, defines, and args can be applied to a "target".
  - Is implicitly found in the directory of a "target", does not need to be included on
    any command line call.

{Colors.green}--------------------------------------------------------------------{Colors.yellow}

  Hello World example:

  The following example is a {Colors.byellow}DEPS.yml{Colors.off}{Colors.yellow} file example for a SystemVerilog simulation of
  hello_world_tb.sv. {Colors.byellow}DEPS.yml{Colors.off}{Colors.yellow} is, in short, a fancy filelist. We use them in the `eda`
  app to organize projects.

--- {Colors.byellow}DEPS.yml{Colors.off}{Colors.yellow}: ---

hello-world:           # <-- this is a named target that will be run

  deps:                # <-- 'deps' is a list of SV, Verilog, VHDL files in compile order
    - hello_world_tb.sv

  top: hello_world_tb  # <-- For testbenches, it is good practice to specifiy the topmost
                       #     module using using 'top'. This is not necessary for design
                       #     files.


--- hello_world_tb.sv: ---

module hello_world_tb;

  initial begin
    #10ns;
    $display("%t %m: Hello World!", $realtime);
    $display("%t %m: Test finished", $realtime);
    $finish;
  end

endmodule : hello_world_tb

---


  hello-world:
    The target name in the {Colors.byellow}DEPS.yml{Colors.off}{Colors.yellow} we named is hello-world. That is a valid target
    that `eda` can use. Such as:

       eda sim --tool=verilator hello-world


{Colors.green}--------------------------------------------------------------------{Colors.yellow}

   Beyond Hello World example:

   The following example is a DEPS.yml file for a more complex module simulation.
   It has two files in ./DEPS.yml and ./lib/DEPS.yml.

--- {Colors.byellow}./DEPS.yml{Colors.off}{Colors.yellow}: ---

my_fifo:                         # <-- this is a design
  incdirs: . lib                 # <-- 'incdirs' define the paths searched to find `include files
  defines:
    FIFO_DEBUG                   # add a basic define
    FIFO_IMPLEMENTATION=uram     # add a define with a value
  deps:            # <-- 'deps' is a list of SV, Verilog, VHDL files in compile order
    - my_fifo.sv                 # an SV file pulled in directly
    - lib/bin_to_gray            # a target, in a subdirectory that has it's own DEPS

my_fifo_test:                    # <-- this is a TEST
  top: my_fifo_test              # the top will default to whatever target is provided
                                 # by the user, so this could be optional
  deps:
    - my_fifo                    # the target that is defined above
    - my_fifo_tb.sv              # an SV file pulled in directly

my_fifo_stress_test:             # <-- this is another TEST
  top: my_fifo_test              # not optional because top is not "my_fifo_stress_test"
  defines:
    STRESS_TEST                  # configures my_fifo_test to be more stressful
  deps:
    - my_fifo_test               # aside from the define, this is same as "my_fifo_test"

--- {Colors.byellow}lib/DEPS.yml{Colors.off}{Colors.yellow}: ---

lib_pkg:                         # <-- this is a package required by bin_to_gray below
  deps:
    - assert_pkg.sv              # an SV package pulled in directly, before it's needed below
    - lib_pkg.sv                 # an SV package pulled in directly

bin_to_gray:                     # <-- this is the target that was required by ../my_fifo
  deps:
    - lib_pkg                    # a target package, listed first as SV requires packages
                                 # to be read before the code that uses them
    - bin_to_gray.sv             # an SV module pulled in directly

{Colors.green}--------------------------------------------------------------------{Colors.yellow}
'''


FULL_DEPS_HELP = f'''

{Colors.green}--------------------------------------------------------------------{Colors.yellow}

''' + get_deps_md_contents()



class CommandDepsHelp(Command):
    '''command handler for: eda deps-help'''

    command_name = 'deps-help'

    def __init__(self, config: dict):
        super().__init__(config)
        # We don't have any args, clear them.
        self.args = {}
        self.args_help = {}
        self.config = config
        self.status = 0

    def process_tokens( # pylint: disable=unused-argument
        self, tokens: list, process_all: bool = True,
        pwd: str = os.getcwd()
    ) -> list:
        '''This is effectively our 'run' method, entrypoint from opencos.eda.main'''

        # disable artifacts, logging is disabled by eda.py adding 'deps-help' to:
        #   util.global_log.default_log_disable_with_args
        util.artifacts.enabled = False

        # run with process_all=False, we don't care about unparsed args
        super().process_tokens(tokens=tokens, process_all=False, pwd=pwd)

        print(BASIC_DEPS_HELP)
        if util.args['verbose'] or util.args['debug']:
            print()
            print(FULL_DEPS_HELP)

        return []

    def help( # pylint: disable=dangerous-default-value
            self, tokens: list = [], no_targets: bool = False,
            show_all_eda_commands: bool = False
    ) -> None:
        '''Since we don't inherit from opencos.eda_base.Command, need our own help
        method
        '''
        super().help(tokens=tokens, no_targets=True)
        print()
        print(BASIC_DEPS_HELP)
        print()
        print(FULL_DEPS_HELP)



''' opencos.eda_tool_helper -- used by pytests and other checks to see if tools are loaded

which helps determine if a pytest is runnable for a given tool, or should be skipped.
Does this without calling `eda` or eda.main(..)

Example uses:
    from opencos import eda_tool_helper
    cfg, tools_loaded = eda_tool_helper.get_config_and_tools_loaded()
    assert 'verilator' in tools_loaded

'''

import os
import shlex
import subprocess
from importlib import import_module
import importlib.util

from opencos import eda_config, util
from opencos.util import Colors
from opencos.utils import vsim_helper, vscode_helper, docker_checks, str_helpers, exe_helper

# Used by pytest, so we can skip tests if tools aren't present, also used by
# opencos.eda itself

DOCKER_TOOL_NOTE = ' [via docker] '

def tool_loaded_but_flagged_as_docker_only(tool: str, config: dict) -> bool:
    '''Checks if a tool was flagged in config as only being able to run

    via docker (and wasn't availble in PATH, shutil.which(..))
    '''
    exe = config.get('auto_tools_found', {}).get(tool, '')
    return exe and exe.startswith(DOCKER_TOOL_NOTE)


def get_config_and_tools_loaded( # pylint: disable=dangerous-default-value
        quiet: bool = False, args: list = [],
        remove_tools_with_docker_note: bool = True
) -> (dict, set):
    '''Returns config dict and list tools_loaded, given the found config.

    Can BYO args such as --config-yml=MY_OWN_EDA_CONFIG.yml
    '''

    # We have to figure out what tools are avaiable w/out calling eda.main,
    # so we can get some of these using eda_config.get_eda_config()
    config, _ = eda_config.get_eda_config(args=args, quiet=quiet)

    # only import 'eda' here so that other methods in this pymodule can be used
    # within eda, etc, if you already have a valid config or tools_loaded.
    eda = import_module("opencos.eda")
    config = eda.init_config(config=config, quiet=quiet)
    tools_loaded = config.get('tools_loaded', []).copy()

    # Remove any tools from tools_loaded that had their exe flagged with
    # str DOCKER_TOOL_NOTE. This is also for pytests so that heplers.can_run_eda_sim()
    # returns False if it were to otherwise appear that we can run verilator
    # (but only with --docker flag set, that majority of test do not)
    if remove_tools_with_docker_note:
        for _tool in list(tools_loaded):
            if tool_loaded_but_flagged_as_docker_only(tool=_tool, config=config):
                tools_loaded.remove(_tool)

    return config, tools_loaded


def get_all_handler_commands( # pylint: disable=too-many-branches
        config=None, tools_loaded=None, auto_tools_only: bool = False
) -> dict:
    '''Given a config and tools_loaded (or if not supplied uses defaults) returns a dict

    of { <command>: [list of tools that run that command, in auto-tool-order] }.

    If you use auto_tools_only=True, then you will only get command: tools that exist in
    config['auto_tools_order'] (and are loaded tools). This would be what you'd get if you
    ran `eda [command]` and did not set the --tool=TOOL arg, for what can run that command.

    If you use auto_tools_only=False (default), you will get the auto tool order +
    tool handlers (loaded tools only) that can service the command. For example, eda waves,
    eda upload, and others may not be in the auto tool list for an `eda [command]`, because
    those command have to self determine their tool.

    For example:
       { "sim": ["verilator", "vivado"],
         "elab": ["slang", "verilator", ...], ...
       }
    '''
    all_handler_commands = {}


    if config and tools_loaded is None:
        tools_loaded = config.get('tools_loaded', [])
    elif config is None or tools_loaded is None:
        # This will re-run eda.init_config(..), b/c you didn't bring a config.
        config, tools_loaded = get_config_and_tools_loaded()


    assert isinstance(config, dict)
    assert isinstance(tools_loaded, list)

    # Let's re-walk auto_tools_order to get this ordered per eda command:
    for command, tools_list in config.get('auto_tools_order', {}).items():

        for tool in tools_list:
            entry = config.get('tools', {}).get(tool, {})
            assert entry, f'{command=} in auto_tools_order {tool=} not present in tools'

            if tool not in tools_loaded:
                continue

            if entry.get('disable-tools-multi', False):
                # Flagged as do-not-add when running eda command: tools-multi
                util.debug(f'eda_tool_helper.py -- skipping {tool=} it is set with flag',
                           'disable-tools-multi in config')
                continue

            if command in entry.get('handlers', {}):
                if command not in all_handler_commands:
                    # create ordered list from config.
                    all_handler_commands[command] = list([tool])
                else:
                    all_handler_commands[command].append(tool)

    if auto_tools_only:
        # skip the next step and return what we've got.
        return all_handler_commands

    # Note that the above will miss commands (waves, upload) that are not in auto_tools_order
    for tool in tools_loaded:
        handlers = config.get('tools', {}).get(tool, {}).get('handlers', {})
        for command, _ in handlers.items():
            if command not in all_handler_commands:
                all_handler_commands[command] = list([tool])
            elif tool not in all_handler_commands.get(command, []):
                all_handler_commands[command].append(tool)

    return all_handler_commands

def get_handler(tool: str, eda_command: str, config: dict, warn: bool = True):
    '''Get the handling class given tool/command/config'''

    entry = config['tools'].get(tool, {})
    if not entry:
        return ''

    handler_name = entry.get('handlers', {}).get(eda_command, '')
    if not handler_name:
        return ''

    module = util.import_class_from_string(handler_name)
    if module:
        return module

    if warn:
        util.warning(
            f'For {tool=}, command {eda_command}, Handler name {handler_name} is',
            'not a valid class to import'
        )
    return None


def get_handler_tool_version(tool: str, eda_command: str, config: dict) -> str:
    '''Attempts to get a Command Handler's version given tool + eda_command'''

    # TODO(drew): maybe if there are no handlers, then do something to get version?
    # like, offer tool_class?

    module = get_handler(tool=tool, eda_command=eda_command, config=config)
    if not module:
        return ''

    obj = module(config=config)

    # Note that Tool.get_versions() is supposed to be 'fast', we don't always
    # run the tool if the 'exe -version' takes too long.
    if getattr(obj, 'get_versions', None):
        return obj.get_versions()

    return ''



def get_handler_info_with_versions( # pylint: disable=too-many-branches,dangerous-default-value
        config: dict = {},
        include_commands: bool = True,
        sort: bool = True,
        show_versions: bool = True,
        all_details: bool = False
) -> str:
    '''Creates and returns a dict of

    {'commands': (what tools/versions can run them),
     'tools': (what version and what commands they can run)
    }

    for arg 'config' you may use:

        config, tools_loaded = get_config_and_tools_loaded()

    show_versions=True can be overriden by config['show_tool_versions']
    all_details=True will make note of Docker and other information.
    '''

    if not config:
        config, tools_loaded = get_config_and_tools_loaded()
    else:
        tools_loaded = list(config.get('tools_loaded', []))

    eda_commands = list(config.get('DEFAULT_HANDLERS', {}).keys())
    if not config.get('show_tool_versions', True):
        # default true, if user set to False in thier config then respect that
        show_versions = False

    info = {
        'tools': {},
    }

    if include_commands:
        info.update({
            'commands': {}
        })
        for eda_command in eda_commands:
            info['commands'][eda_command] = {} # init

    # set tool version, and commands supported
    for tool in tools_loaded:

        info['tools'][tool] = {}
        if show_versions:
            info['tools'][tool]['version'] = ''
        if include_commands:
            info['tools'][tool]['commands'] = []

        for eda_command in eda_commands:

            ver = ''
            if not show_versions:
                # well then we also don't know if the tool is loaded?
                if not include_commands:
                    continue
                if not get_handler(tool=tool, eda_command=eda_command, config=config,
                                   warn=False):
                    continue

            else:
                if not include_commands and info['tools'][tool]['version']:
                    # version is already set, and we're not doing for all commands:
                    break

                # Note that if you have a generic handler, or a tool that has
                # several handlers with more than one Tool class, that all can return a
                # non blank-str version, you may have problems.
                ver = get_handler_tool_version(
                    tool=tool, eda_command=eda_command, config=config
                )

                if not ver:
                    continue

            info['tools'][tool]['version'] = ver

            if include_commands:
                info['commands'][eda_command][tool] = ver
                info['tools'][tool]['commands'].append(eda_command)

    # set path, and support all_details=True
    for tool, _ in info['tools'].items():
        if path := config.get('auto_tools_found', {}).get(tool, ''):
            info['tools'][tool]['path'] = path

            if all_details:
                if path.startswith(DOCKER_TOOL_NOTE):
                    info['tools'][tool]['docker'] = {
                        'image': path.replace(DOCKER_TOOL_NOTE, '')
                    }

    # return the info dict with 'tools' and 'commands' entries sorted:
    if sort:
        for key in list(info.keys()):
            if info[key]:
                info[key] = dict(sorted(info[key].items()))

    return info


def pretty_info_handler_tools( # pylint: disable=dangerous-default-value
        info: dict = {}, config: dict = {}, command: str = ''
) -> None:
    '''Pretty print (via util.info) the result from get_handler_info_with_versions()

    if info is None or empty, will use config to run get_handler_info_with_versions(..)

    Does not include commands
    '''

    if not info:
        info = get_handler_info_with_versions(config=config, include_commands=False, sort=True)

    if not info.get('tools', {}):
        # No tools detected
        if command and command in config.get('command_tool_is_optional', []):
            # but this command doesn't need tools
            return
        if config.get('ssh', None):
            # For ssh we don't require tools, will be sent to remote, avoid this message.
            return

        # if command omitted, or command may need tools, print that we don't have any
        util.info('No tools detected!', color=Colors.yellow)
        return

    show_versions = any(dvalue.get('version', '') for dvalue in info['tools'].values())

    tools_rows = [
        ['--Detected tool--', '--Path--'] # Header row.
    ]

    if show_versions:
        tools_rows[0].append('--Version--')

    for _tool, dvalue in info['tools'].items():
        path = dvalue.get("path", "")
        if path:
            path = f'({path})'

        if show_versions:
            version = dvalue.get("version", "")
            # will defer printing again, so we can put them into aligned columns:
            tools_rows.append([_tool, path, version])
        else:
            tools_rows.append([_tool, path])

    # Since Docker isn't technically a tool, and only lives in opencos.utils.docker_checks,
    # we would like to still list it here (as "other") to at least show path/version information.
    # This also makes it less surprising if an actual tool has their PATH set with ' [via docker] '
    # (DOCKER_TOOL_NOTE str)
    other_rows = [
        ['--Detected other--', '--Path--']
    ]
    if show_versions:
        other_rows[0].append('--Version--')
    if docker_checks.DOCKER_EXE:
        docker_entry = ['docker', docker_checks.DOCKER_EXE]
        if show_versions:
            docker_entry.append(docker_checks.get_docker_version())
        other_rows.append(docker_entry)

    # Finally, print everything
    pretty_info_tool_rows(tools_rows + other_rows)


def pretty_info_tool_rows(rows: list) -> None:
    '''Pretty print a 2-d list, called by pretty_info_handler_tools'''

    for rownum, row in enumerate(str_helpers.pretty_2dlist_columns(
            rows, return_as_2d_list=True, header_row_centered=False)):
        if rownum == 0 or row[0].startswith('--'):
            util.info("".join(row), color=Colors.bgreen)
        else:
            # get the results in a padded 2D list so we can colorize the name (index 0)
            util.info(f'{row[0]}{Colors.cyan}' \
                      + ''.join(row[1:]), color=Colors.bcyan)


def check_config_tool_requirements( # pylint: disable=too-many-locals,too-many-branches,too-many-statements
        config: dict, tool: str
) -> dict:
    '''Returns a dict of key: bool for various checks that are

    required for this tool in config.tools.TOOLNAME. If tool is not
    present in config, it will return an empty dict.
    '''
    ret = {
        'exe': '',
        'exe_list': [],
        'ALL_CHECKS': False,
    }

    tool_cfg = config.get('tools', {}).get(tool, {})
    if not tool_cfg:
        return ret

    exe_list = eda_config.tool_get_exe_list(tool=tool, config=config)
    ret.update({
        'exe': exe_list[0],
        'exe_list': exe_list
    })

    requires_py_list = tool_cfg.get('requires-py', [])
    for pkg in requires_py_list:
        spec = importlib.util.find_spec(pkg)
        if not spec:
            ret[f'has-required-py__{pkg}'] = False
            util.debug(f"... No, missing python pkg {pkg}: {spec=}")

    requires_env_list = tool_cfg.get('requires-env', [])
    for env in requires_env_list:
        if not os.environ.get(env, ''):
            ret[f'has-required-env__{env}'] = False
            util.debug(f"... No, missing env {env}")

    exe_fpath = None
    for exe in exe_list:
        if exe is not None:
            found_path = exe_helper.get_tool_exe(config=config, tool=tool)
            if not exe_fpath:
                exe_fpath = found_path
                ret['exe'] = found_path
        else:
            found_path = None
        if not found_path:
            # skip the fail if we can run via docker instead:
            if tool_cfg.get('docker-support', None) or \
               tool_cfg.get('requires-docker', None):
                if not docker_checks.docker_ok():
                    ret[f'has-required-exe__{exe}'] = False
                    ret[f'has-required-exe__{exe}__no-docker'] = False
                else:
                    ret[f'has-required-exe__{exe}'] = True
                    ret[f'has-required-exe__{exe}__docker-ok'] = True
                    ret['exe'] = DOCKER_TOOL_NOTE
            else:
                ret[f'has-required-exe__{exe}'] = False

    docker_default_image = tool_cfg.get('docker-default-image', '')
    if docker_checks.docker_ok() and ret['exe'].startswith(DOCKER_TOOL_NOTE) and \
       docker_default_image:
        # We could make sure the docker-default-image exists on this system, and
        # bonus points if we also pre-parsed any --docker-image arg to check that
        # too. However, the goal here isn't to say "you can't run this tool", we
        # simply don't want to force the user to have to pull images they never
        # asked for.
        # Instead, simply show the exe as: " [ via docker ] IMAGE"
        ret['exe'] += docker_default_image


    # has_vsim_helper is handled by exe_helper.get_tool_exe(config, tool)
    # but we'll run it again for debug information:
    if tool_cfg.get('requires-vsim-helper', False):
        # This tool name must be in opencos.utils.vsim_helper.TOOL_PATH[tool].
        # Special case for vsim being used by a lot of tools.
        vsim_helper.init() # only runs checks once internally
        found_path = vsim_helper.TOOL_PATH[tool]
        if not found_path:
            ret['has-required-vsim-helper'] = False

    needs_vscode_extensions = tool_cfg.get('requires-vscode-extension', None)
    if needs_vscode_extensions:
        if not isinstance(needs_vscode_extensions, list):
            util.error(
                f'eda config issue, tool {tool}: requires-vscode-extension must be a list'
            )
        else:
            vscode_helper.init() # only runs checks once internally
            for x in needs_vscode_extensions:
                ret[f'has-required-vscode-extension__{x}'] = (
                    x in vscode_helper.EXTENSIONS
                )

    if ret['exe'] or ret['exe'] is None:
        # We'll run this if we found something for the tools.TOOL.exe,
        # OR, if tools.TOOL.exe is None, means we're likely in a situation where
        # docker-required was also True
        requires_cmd_list = tool_cfg.get('requires-cmd', [])
        for i, cmd in enumerate(requires_cmd_list):
            cmd_list = shlex.split(cmd)
            try:
                proc = subprocess.run(
                    cmd_list, capture_output=True, check=False, input=b'exit\n\n'
                )
                if proc.returncode != 0:
                    util.debug(f"For tool {tool} missing required command {i=}",
                               f"({proc.returncode=}): {cmd_list=}")
                    ret[f'has-required-cmd__{i}'] = False
            except Exception as e:
                ret[f'has-required-cmd__{i}'] = False
                util.debug(f"... No, exception {e} running requires-cmd [{i}]: {cmd_list}")

    needs_docker = tool_cfg.get('requires-docker', None)
    if needs_docker:
        # Confirm that we either have an exe, or that it's set to DOCKER_TOOL_NOTE.
        exe = ret['exe_list'][0]
        ret['has-required-docker'] = docker_checks.docker_ok()


    del ret['ALL_CHECKS']
    ret.update({
        'ALL_CHECKS': all((bool(x) for x in ret.values()))
    })

    return ret
